char TCL_LIBS[] = "\
# init.tcl --\n\
#\n\
# Default system startup file for Tcl-based applications.  Defines\n\
# \"unknown\" procedure and auto-load facilities.\n\
#\n\
# $Header: /user6/ouster/tcl/library/RCS/init.tcl,v 1.28 93/10/08 09:11:21 ouster Exp $ SPRITE (Berkeley)\n\
#\n\
# Copyright (c) 1991-1993 The Regents of the University of California.\n\
# All rights reserved.\n\
#\n\
# Permission is hereby granted, without written agreement and without\n\
# license or royalty fees, to use, copy, modify, and distribute this\n\
# software and its documentation for any purpose, provided that the\n\
# above copyright notice and the following two paragraphs appear in\n\
# all copies of this software.\n\
#\n\
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR\n\
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT\n\
# OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF\n\
# CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n\
#\n\
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,\n\
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY\n\
# AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS\n\
# ON AN \"AS IS\" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO\n\
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.\n\
#\n\
\n\
set auto_path [info library]\n\
\n\
# unknown:\n\
# Invoked when a Tcl command is invoked that doesn't exist in the\n\
# interpreter:\n\
#\n\
#	1. See if the autoload facility can locate the command in a\n\
#	   Tcl script file.  If so, load it and execute it.\n\
#	2. See if the command exists as an executable UNIX program.\n\
#	   If so, \"exec\" the command.\n\
#	3. If the command was invoked at top-level:\n\
#	    (a) see if the command requests csh-like history substitution\n\
#		in one of the common forms !!, !<number>, or ^old^new.  If\n\
#		so, emulate csh's history substitution.\n\
#	    (b) see if the command is a unique abbreviation for another\n\
#		command.  If so, invoke the command.\n\
\n\
proc unknown args {\n\
    global auto_noexec auto_noload env unknown_pending tcl_interactive;\n\
\n\
    set name [lindex $args 0]\n\
    if ![info exists auto_noload] {\n\
	#\n\
	# Make sure we're not trying to load the same proc twice.\n\
	#\n\
	if [info exists unknown_pending($name)] {\n\
	    unset unknown_pending($name)\n\
	    if {[array size unknown_pending] == 0} {\n\
		unset unknown_pending\n\
	    }\n\
	    return -code error \"self-referential recursion in \\\"unknown\\\" for command \\\"$name\\\"\";\n\
	}\n\
	set unknown_pending($name) pending;\n\
	set ret [catch {auto_load $name} msg]\n\
	unset unknown_pending($name);\n\
	if {$ret != 0} {\n\
	    return -code $ret \"error while autoloading \\\"$name\\\": $msg\"\n\
	}\n\
	if ![array size unknown_pending] {\n\
	    unset unknown_pending\n\
	}\n\
	if $msg {\n\
	    return [uplevel $args]\n\
	}\n\
    }\n\
    if {([info level] == 1) && ([info script] == \"\") && $tcl_interactive} {\n\
	if ![info exists auto_noexec] {\n\
	    if [auto_execok $name] {\n\
		return [uplevel exec >&@stdout <@stdin $args]\n\
	    }\n\
	}\n\
	if {$name == \"!!\"} {\n\
	    return [uplevel {history redo}]\n\
	}\n\
	if [regexp {^!(.+)$} $name dummy event] {\n\
	    return [uplevel [list history redo $event]]\n\
	}\n\
	if [regexp {^\\^([^^]*)\\^([^^]*)\\^?$} $name dummy old new] {\n\
	    return [uplevel [list history substitute $old $new]]\n\
	}\n\
	set cmds [info commands $name*]\n\
	if {[llength $cmds] == 1} {\n\
	    return [uplevel [lreplace $args 0 0 $cmds]]\n\
	}\n\
	if {[llength $cmds] != 0} {\n\
	    if {$name == \"\"} {\n\
		return -code error \"empty command name \\\"\\\"\"\n\
	    } else {\n\
		return -code error \\\n\
			\"ambiguous command name \\\"$name\\\": [lsort $cmds]\"\n\
	    }\n\
	}\n\
    }\n\
    return -code error \"invalid command name \\\"$name\\\"\"\n\
}\n\
\n\
# auto_load:\n\
# Checks a collection of library directories to see if a procedure\n\
# is defined in one of them.  If so, it sources the appropriate\n\
# library file to create the procedure.  Returns 1 if it successfully\n\
# loaded the procedure, 0 otherwise.\n\
\n\
proc auto_load cmd {\n\
    global auto_index auto_oldpath auto_path env errorInfo errorCode\n\
\n\
    if [info exists auto_index($cmd)] {\n\
	uplevel #0 $auto_index($cmd)\n\
	return 1\n\
    }\n\
    if [catch {set path $auto_path}] {\n\
	if [catch {set path $env(TCLLIBPATH)}] {\n\
	    if [catch {set path [info library]}] {\n\
		return 0\n\
	    }\n\
	}\n\
    }\n\
    if [info exists auto_oldpath] {\n\
	if {$auto_oldpath == $path} {\n\
	    return 0\n\
	}\n\
    }\n\
    set auto_oldpath $path\n\
    catch {unset auto_index}\n\
    for {set i [expr [llength $path] - 1]} {$i >= 0} {incr i -1} {\n\
	set dir [lindex $path $i]\n\
	set f \"\"\n\
	if [catch {set f [open $dir/tclIndex]}] {\n\
	    continue\n\
	}\n\
	set error [catch {\n\
	    set id [gets $f]\n\
	    if {$id == \"# Tcl autoload index file, version 2.0\"} {\n\
		eval [read $f]\n\
	    } elseif {$id == \"# Tcl autoload index file: each line identifies a Tcl\"} {\n\
		while {[gets $f line] >= 0} {\n\
		    if {([string index $line 0] == \"#\")\n\
			    || ([llength $line] != 2)} {\n\
			continue\n\
		    }\n\
		    set name [lindex $line 0]\n\
		    set auto_index($name) \"source $dir/[lindex $line 1]\"\n\
		}\n\
	    } else {\n\
		error \"$dir/tclIndex isn't a proper Tcl index file\"\n\
	    }\n\
	} msg]\n\
	if {$f != \"\"} {\n\
	    close $f\n\
	}\n\
	if $error {\n\
	    error $msg $errorInfo $errorCode\n\
	}\n\
    }\n\
    if [info exists auto_index($cmd)] {\n\
	uplevel #0 $auto_index($cmd)\n\
	if {[info commands $cmd] != \"\"} {\n\
	    return 1\n\
	}\n\
    }\n\
    return 0\n\
}\n\
\n\
# auto_execok:\n\
# Returns 1 if there's an executable in the current path for the\n\
# given name, 0 otherwise.  Builds an associative array auto_execs\n\
# that caches information about previous checks, for speed.\n\
\n\
proc auto_execok name {\n\
    global auto_execs env\n\
\n\
    if [info exists auto_execs($name)] {\n\
	return $auto_execs($name)\n\
    }\n\
    set auto_execs($name) 0\n\
    if {[string first / $name] >= 0} {\n\
	if {[file executable $name] && ![file isdirectory $name]} {\n\
	    set auto_execs($name) 1\n\
	}\n\
	return $auto_execs($name)\n\
    }\n\
    foreach dir [split $env(PATH) :] {\n\
	if {[file executable $dir/$name] && ![file isdirectory $dir/$name]} {\n\
	    set auto_execs($name) 1\n\
	    return 1\n\
	}\n\
    }\n\
    return 0\n\
}\n\
\n\
# auto_reset:\n\
# Destroy all cached information for auto-loading and auto-execution,\n\
# so that the information gets recomputed the next time it's needed.\n\
# Also delete any procedures that are listed in the auto-load index\n\
# except those related to auto-loading.\n\
\n\
proc auto_reset {} {\n\
    global auto_execs auto_index auto_oldpath\n\
    foreach p [info procs] {\n\
	if {[info exists auto_index($p)] && ($p != \"unknown\")\n\
		&& ![string match auto_* $p]} {\n\
	    rename $p {}\n\
	}\n\
    }\n\
    catch {unset auto_execs}\n\
    catch {unset auto_index}\n\
    catch {unset auto_oldpath}\n\
}\n\
\n\
# auto_mkindex:\n\
# Regenerate a tclIndex file from Tcl source files.  Takes as argument\n\
# the name of the directory in which the tclIndex file is to be placed,\n\
# floowed by any number of glob patterns to use in that directory to\n\
# locate all of the relevant files.\n\
\n\
proc auto_mkindex {dir args} {\n\
    global errorCode errorInfo\n\
    set oldDir [pwd]\n\
    cd $dir\n\
    set dir [pwd]\n\
    append index \"# Tcl autoload index file, version 2.0\\n\"\n\
    append index \"# This file is generated by the \\\"auto_mkindex\\\" command\\n\"\n\
    append index \"# and sourced to set up indexing information for one or\\n\"\n\
    append index \"# more commands.  Typically each line is a command that\\n\"\n\
    append index \"# sets an element in the auto_index array, where the\\n\"\n\
    append index \"# element name is the name of a command and the value is\\n\"\n\
    append index \"# a script that loads the command.\\n\\n\"\n\
    foreach file [eval glob $args] {\n\
	set f \"\"\n\
	set error [catch {\n\
	    set f [open $file]\n\
	    while {[gets $f line] >= 0} {\n\
		if [regexp {^proc[ 	]+([^ 	]*)} $line match procName] {\n\
		    append index \"set [list auto_index($procName)]\"\n\
		    append index \" \\\"source \\$dir/$file\\\"\\n\"\n\
		}\n\
	    }\n\
	    close $f\n\
	} msg]\n\
	if $error {\n\
	    set code $errorCode\n\
	    set info $errorInfo\n\
	    catch {close $f}\n\
	    cd $oldDir\n\
	    error $msg $info $code\n\
	}\n\
    }\n\
    set f [open tclIndex w]\n\
    puts $f $index nonewline\n\
    close $f\n\
    cd $oldDir\n\
}\n\
# parray:\n\
# Print the contents of a global array on stdout.\n\
#\n\
# $Header: /user6/ouster/tcl/library/RCS/parray.tcl,v 1.5 93/02/06 16:33:45 ouster Exp $ SPRITE (Berkeley)\n\
#\n\
# Copyright (c) 1991-1993 The Regents of the University of California.\n\
# All rights reserved.\n\
#\n\
# Permission is hereby granted, without written agreement and without\n\
# license or royalty fees, to use, copy, modify, and distribute this\n\
# software and its documentation for any purpose, provided that the\n\
# above copyright notice and the following two paragraphs appear in\n\
# all copies of this software.\n\
#\n\
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR\n\
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT\n\
# OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF\n\
# CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n\
#\n\
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,\n\
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY\n\
# AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS\n\
# ON AN \"AS IS\" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO\n\
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.\n\
#\n\
\n\
proc parray a {\n\
    upvar 1 $a array\n\
    if [catch {array size array}] {\n\
	error \"\\\"$a\\\" isn't an array\"\n\
    }\n\
    set maxl 0\n\
    foreach name [lsort [array names array]] {\n\
	if {[string length $name] > $maxl} {\n\
	    set maxl [string length $name]\n\
	}\n\
    }\n\
    set maxl [expr {$maxl + [string length $a] + 2}]\n\
    foreach name [lsort [array names array]] {\n\
	set nameString [format %s(%s) $a $name]\n\
	puts stdout [format \"%-*s = %s\" $maxl $nameString $array($name)]\n\
    }\n\
}\n\
# button.tcl --\n\
#\n\
# This file contains Tcl procedures used to manage Tk buttons.\n\
#\n\
# $Header: /user6/ouster/wish/library/RCS/button.tcl,v 1.9 93/07/01 13:41:53 ouster Exp $ SPRITE (Berkeley)\n\
#\n\
# Copyright (c) 1992-1993 The Regents of the University of California.\n\
# All rights reserved.\n\
#\n\
# Permission is hereby granted, without written agreement and without\n\
# license or royalty fees, to use, copy, modify, and distribute this\n\
# software and its documentation for any purpose, provided that the\n\
# above copyright notice and the following two paragraphs appear in\n\
# all copies of this software.\n\
#\n\
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR\n\
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT\n\
# OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF\n\
# CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n\
#\n\
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,\n\
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY\n\
# AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS\n\
# ON AN \"AS IS\" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO\n\
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.\n\
#\n\
\n\
# The procedure below is invoked when the mouse pointer enters a\n\
# button widget.  It records the button we're in and changes the\n\
# state of the button to active unless the button is disabled.\n\
\n\
proc tk_butEnter w {\n\
    global tk_priv tk_strictMotif\n\
    if {[lindex [$w config -state] 4] != \"disabled\"} {\n\
	if {!$tk_strictMotif} {\n\
	    $w config -state active\n\
	}\n\
	set tk_priv(window) $w\n\
    }\n\
}\n\
\n\
# The procedure below is invoked when the mouse pointer leaves a\n\
# button widget.  It changes the state of the button back to\n\
# inactive.\n\
\n\
proc tk_butLeave w {\n\
    global tk_priv tk_strictMotif\n\
    if {[lindex [$w config -state] 4] != \"disabled\"} {\n\
	if {!$tk_strictMotif} {\n\
	    $w config -state normal\n\
	}\n\
    }\n\
    set tk_priv(window) \"\"\n\
}\n\
\n\
# The procedure below is invoked when the mouse button is pressed in\n\
# a button/radiobutton/checkbutton widget.  It records information\n\
# (a) to indicate that the mouse is in the button, and\n\
# (b) to save the button's relief so it can be restored later.\n\
\n\
proc tk_butDown w {\n\
    global tk_priv\n\
    set tk_priv(relief) [lindex [$w config -relief] 4]\n\
    set tk_priv(buttonWindow) $w\n\
    if {[lindex [$w config -state] 4] != \"disabled\"} {\n\
	$w config -relief sunken\n\
    }\n\
}\n\
\n\
# The procedure below is invoked when the mouse button is released\n\
# for a button/radiobutton/checkbutton widget.  It restores the\n\
# button's relief and invokes the command as long as the mouse\n\
# hasn't left the button.\n\
\n\
proc tk_butUp w {\n\
    global tk_priv\n\
    if {$w == $tk_priv(buttonWindow)} {\n\
	$w config -relief $tk_priv(relief)\n\
	if {($w == $tk_priv(window))\n\
		&& ([lindex [$w config -state] 4] != \"disabled\")} {\n\
	    uplevel #0 [list $w invoke]\n\
	}\n\
	set tk_priv(buttonWindow) \"\"\n\
    }\n\
}\n\
# dialog.tcl --\n\
#\n\
# This file defines the procedure tk_dialog, which creates a dialog\n\
# box containing a bitmap, a message, and one or more buttons.\n\
#\n\
# $Header: /user6/ouster/wish/library/RCS/dialog.tcl,v 1.4 93/08/16 16:59:52 ouster Exp $ SPRITE (Berkeley)\n\
#\n\
# Copyright (c) 1992-1993 The Regents of the University of California.\n\
# All rights reserved.\n\
#\n\
# Permission is hereby granted, without written agreement and without\n\
# license or royalty fees, to use, copy, modify, and distribute this\n\
# software and its documentation for any purpose, provided that the\n\
# above copyright notice and the following two paragraphs appear in\n\
# all copies of this software.\n\
#\n\
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR\n\
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT\n\
# OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF\n\
# CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n\
#\n\
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,\n\
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY\n\
# AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS\n\
# ON AN \"AS IS\" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO\n\
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.\n\
#\n\
\n\
#\n\
# tk_dialog:\n\
#\n\
# This procedure displays a dialog box, waits for a button in the dialog\n\
# to be invoked, then returns the index of the selected button.\n\
#\n\
# Arguments:\n\
# w -		Window to use for dialog top-level.\n\
# title -	Title to display in dialog's decorative frame.\n\
# text -	Message to display in dialog.\n\
# bitmap -	Bitmap to display in dialog (empty string means none).\n\
# default -	Index of button that is to display the default ring\n\
#		(-1 means none).\n\
# args -	One or more strings to display in buttons across the\n\
#		bottom of the dialog box.\n\
\n\
proc tk_dialog {w title text bitmap default args} {\n\
    global tk_priv\n\
\n\
    # 1. Create the top-level window and divide it into top\n\
    # and bottom parts.\n\
\n\
    catch {destroy $w}\n\
    toplevel $w -class Dialog\n\
    wm title $w $title\n\
    wm iconname $w Dialog\n\
    frame $w.top -relief raised -bd 1\n\
    pack $w.top -side top -fill both\n\
    frame $w.bot -relief raised -bd 1\n\
    pack $w.bot -side bottom -fill both\n\
\n\
    # 2. Fill the top part with bitmap and message.\n\
\n\
    message $w.msg -width 3i -text $text \\\n\
	    -font -Adobe-Times-Medium-R-Normal-*-180-*\n\
    pack $w.msg -in $w.top -side right -expand 1 -fill both -padx 5m -pady 5m\n\
    if {$bitmap != \"\"} {\n\
	label $w.bitmap -bitmap $bitmap\n\
	pack $w.bitmap -in $w.top -side left -padx 5m -pady 5m\n\
    }\n\
\n\
    # 3. Create a row of buttons at the bottom of the dialog.\n\
\n\
    set i 0\n\
    foreach but $args {\n\
	button $w.button$i -text $but -command \"set tk_priv(button) $i\"\n\
	if {$i == $default} {\n\
	    frame $w.default -relief sunken -bd 1\n\
	    raise $w.button$i $w.default\n\
	    pack $w.default -in $w.bot -side left -expand 1 -padx 3m -pady 2m\n\
	    pack $w.button$i -in $w.default -padx 2m -pady 2m \\\n\
		    -ipadx 2m -ipady 1m\n\
	    bind $w <Return> \"$w.button$i flash; set tk_priv(button) $i\"\n\
	} else {\n\
	    pack $w.button$i -in $w.bot -side left -expand 1 \\\n\
		    -padx 3m -pady 3m -ipadx 2m -ipady 1m\n\
	}\n\
	incr i\n\
    }\n\
\n\
    # 4. Withdraw the window, then update all the geometry information\n\
    # so we know how big it wants to be, then center the window in the\n\
    # display and de-iconify it.\n\
\n\
    wm withdraw $w\n\
    update idletasks\n\
    set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \\\n\
	    - [winfo vrootx [winfo parent $w]]]\n\
    set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 \\\n\
	    - [winfo vrooty [winfo parent $w]]]\n\
    wm geom $w +$x+$y\n\
    wm deiconify $w\n\
\n\
    # 5. Set a grab and claim the focus too.\n\
\n\
    set oldFocus [focus]\n\
    grab $w\n\
    focus $w\n\
\n\
    # 6. Wait for the user to respond, then restore the focus and\n\
    # return the index of the selected button.\n\
\n\
    tkwait variable tk_priv(button)\n\
    destroy $w\n\
    focus $oldFocus\n\
    return $tk_priv(button)\n\
}\n\
# entry.tcl --\n\
#\n\
# This file contains Tcl procedures used to manage Tk entries.\n\
#\n\
# $Header: /user6/ouster/wish/library/RCS/entry.tcl,v 1.7 93/10/18 17:15:23 ouster Exp $ SPRITE (Berkeley)\n\
#\n\
# Copyright (c) 1992-1993 The Regents of the University of California.\n\
# All rights reserved.\n\
#\n\
# Permission is hereby granted, without written agreement and without\n\
# license or royalty fees, to use, copy, modify, and distribute this\n\
# software and its documentation for any purpose, provided that the\n\
# above copyright notice and the following two paragraphs appear in\n\
# all copies of this software.\n\
#\n\
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR\n\
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT\n\
# OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF\n\
# CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n\
#\n\
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,\n\
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY\n\
# AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS\n\
# ON AN \"AS IS\" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO\n\
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.\n\
#\n\
\n\
# The procedure below is invoked to backspace over one character\n\
# in an entry widget.  The name of the widget is passed as argument.\n\
\n\
proc tk_entryBackspace w {\n\
    set x [expr {[$w index insert] - 1}]\n\
    if {$x != -1} {$w delete $x}\n\
}\n\
\n\
# The procedure below is invoked to backspace over one word in an\n\
# entry widget.  The name of the widget is passed as argument.\n\
\n\
proc tk_entryBackword w {\n\
    set string [$w get]\n\
    set curs [expr [$w index insert]-1]\n\
    if {$curs < 0} return\n\
    for {set x $curs} {$x > 0} {incr x -1} {\n\
	if {([string first [string index $string $x] \" \\t\"] < 0)\n\
		&& ([string first [string index $string [expr $x-1]] \" \\t\"]\n\
		>= 0)} {\n\
	    break\n\
	}\n\
    }\n\
    $w delete $x $curs\n\
}\n\
\n\
# The procedure below is invoked after insertions.  If the caret is not\n\
# visible in the window then the procedure adjusts the entry's view to\n\
# bring the caret back into the window again.  Also, try to keep at\n\
# least one character visible to the left of the caret.\n\
\n\
proc tk_entrySeeCaret w {\n\
    set c [$w index insert]\n\
    set left [$w index @0]\n\
    if {$left >= $c} {\n\
	if {$c > 0} {\n\
	    $w view [expr $c-1]\n\
	} else {\n\
	    $w view $c\n\
	}\n\
	return\n\
    }\n\
    set x [expr [winfo width $w] - [lindex [$w config -bd] 4] - 1]\n\
    while {([$w index @$x] < $c) && ($left < $c)} {\n\
	set left [expr $left+1]\n\
	$w view $left\n\
    }\n\
}\n\
# listbox.tcl --\n\
#\n\
# This file contains Tcl procedures used to manage Tk listboxes.\n\
#\n\
# $Header: /user6/ouster/wish/library/RCS/listbox.tcl,v 1.3 93/07/01 13:42:05 ouster Exp $ SPRITE (Berkeley)\n\
#\n\
# Copyright (c) 1992-1993 The Regents of the University of California.\n\
# All rights reserved.\n\
#\n\
# Permission is hereby granted, without written agreement and without\n\
# license or royalty fees, to use, copy, modify, and distribute this\n\
# software and its documentation for any purpose, provided that the\n\
# above copyright notice and the following two paragraphs appear in\n\
# all copies of this software.\n\
#\n\
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR\n\
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT\n\
# OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF\n\
# CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n\
#\n\
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,\n\
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY\n\
# AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS\n\
# ON AN \"AS IS\" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO\n\
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.\n\
#\n\
\n\
# The procedure below may be invoked to change the behavior of\n\
# listboxes so that only a single item may be selected at once.\n\
# The arguments give one or more windows whose behavior should\n\
# be changed;  if one of the arguments is \"Listbox\" then the default\n\
# behavior is changed for all listboxes.\n\
\n\
proc tk_listboxSingleSelect args {\n\
    foreach w $args {\n\
	bind $w <B1-Motion> {%W select from [%W nearest %y]} \n\
	bind $w <Shift-1> {%W select from [%W nearest %y]}\n\
	bind $w <Shift-B1-Motion> {%W select from [%W nearest %y]}\n\
    }\n\
}\n\
# menu.tcl --\n\
#\n\
# This file contains Tcl procedures used to manage Tk menus and\n\
# menubuttons.  Most of the code here is dedicated to support for\n\
# pulling down menus and menu traversal via the keyboard.\n\
#\n\
# $Header: /user6/ouster/wish/library/RCS/menu.tcl,v 1.23 93/09/17 14:02:28 ouster Exp $ SPRITE (Berkeley)\n\
#\n\
# Copyright (c) 1992-1993 The Regents of the University of California.\n\
# All rights reserved.\n\
#\n\
# Permission is hereby granted, without written agreement and without\n\
# license or royalty fees, to use, copy, modify, and distribute this\n\
# software and its documentation for any purpose, provided that the\n\
# above copyright notice and the following two paragraphs appear in\n\
# all copies of this software.\n\
#\n\
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR\n\
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT\n\
# OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF\n\
# CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n\
#\n\
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,\n\
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY\n\
# AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS\n\
# ON AN \"AS IS\" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO\n\
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.\n\
#\n\
\n\
# The procedure below is publically available.  It is used to identify\n\
# a frame that serves as a menu bar and the menu buttons that lie inside\n\
# the menu bar.  This procedure establishes proper \"menu bar\" behavior\n\
# for all of the menu buttons, including keyboard menu traversal.  Only\n\
# one menu bar may exist for a given top-level window at a time.\n\
# Arguments:\n\
#	\n\
# bar -				The path name of the containing frame.  Must\n\
#				be an ancestor of all of the menu buttons,\n\
#				since it will be be used in grabs.\n\
# additional arguments -	One or more menu buttons that are descendants\n\
#				of bar.  The order of these arguments\n\
#				determines the order of keyboard traversal.\n\
#				If no extra arguments are named then all of\n\
#				the menu bar information for bar is cancelled.\n\
\n\
proc tk_menuBar {w args} {\n\
    global tk_priv\n\
    if {$args == \"\"} {\n\
	if [catch {set menus $tk_priv(menusFor$w)}] {\n\
	    return \"\"\n\
	}\n\
	return $menus\n\
    }\n\
    if [info exists tk_priv(menusFor$w)] {\n\
	unset tk_priv(menusFor$w)\n\
	unset tk_priv(menuBarFor[winfo toplevel $w])\n\
    }\n\
    if {$args == \"{}\"} {\n\
	return\n\
    }\n\
    set tk_priv(menusFor$w) $args\n\
    set tk_priv(menuBarFor[winfo toplevel $w]) $w\n\
    bind $w <Any-Alt-KeyPress> {tk_traverseToMenu %W %A}\n\
    bind $w <F10> {tk_firstMenu %W}\n\
    bind $w <Any-ButtonRelease-1> tk_mbUnpost\n\
}\n\
\n\
proc tk_menus {w args} {\n\
    error \"tk_menus is obsolete in Tk versions 3.0 and later; please change your scripts to use tk_menuBar instead\"\n\
}\n\
\n\
# The procedure below is publically available.  It takes any number of\n\
# arguments that are names of widgets or classes.  It sets up bindings\n\
# for the widgets or classes so that keyboard menu traversal is possible\n\
# when the input focus is in those widgets or classes.\n\
\n\
proc tk_bindForTraversal args {\n\
    foreach w $args {\n\
	bind $w <Any-Alt-KeyPress> {tk_traverseToMenu %W %A}\n\
	bind $w <F10> {tk_firstMenu %W}\n\
    }\n\
}\n\
\n\
# The procedure below does all of the work of posting a menu (including\n\
# unposting any other menu that might currently be posted).  The \"w\"\n\
# argument is the name of the menubutton for the menu to be posted.\n\
# Note:  if $w is disabled then the procedure does nothing.\n\
\n\
proc tk_mbPost {w} {\n\
    global tk_priv tk_strictMotif\n\
    if {[lindex [$w config -state] 4] == \"disabled\"} {\n\
	return\n\
    }\n\
    if {$w == $tk_priv(posted)} {\n\
	grab -global $tk_priv(grab)\n\
	return\n\
    }\n\
    set menu [lindex [$w config -menu] 4]\n\
    if {$menu == \"\"} {\n\
	return\n\
    }\n\
    if ![string match $w* $menu] {\n\
	error \"can't post $menu:  it isn't a descendant of $w (this is a new requirement in Tk versions 3.0 and later)\"\n\
    }\n\
    set cur $tk_priv(posted)\n\
    if {$cur != \"\"} tk_mbUnpost\n\
    set tk_priv(relief) [lindex [$w config -relief] 4]\n\
    $w config -relief raised\n\
    set tk_priv(posted) $w\n\
    if {$tk_priv(focus) == \"\"} {\n\
	set tk_priv(focus) [focus]\n\
    }\n\
    set tk_priv(activeBg) [lindex [$menu config -activebackground] 4]\n\
    set tk_priv(activeFg) [lindex [$menu config -activeforeground] 4]\n\
    if $tk_strictMotif {\n\
	$menu config -activebackground [lindex [$menu config -background] 4]\n\
	$menu config -activeforeground [lindex [$menu config -foreground] 4]\n\
    }\n\
    $menu activate none\n\
    focus $menu\n\
    $menu post [winfo rootx $w] [expr [winfo rooty $w]+[winfo height $w]]\n\
    if [catch {set grab $tk_priv(menuBarFor[winfo toplevel $w])}] {\n\
	set grab $w\n\
    } else {\n\
	if [lsearch $tk_priv(menusFor$grab) $w]<0 {\n\
	    set grab $w\n\
	}\n\
    }\n\
    set tk_priv(cursor) [lindex [$grab config -cursor] 4]\n\
    $grab config -cursor arrow\n\
    set tk_priv(grab) $grab\n\
    grab -global $grab\n\
}\n\
\n\
# The procedure below does all the work of unposting the menubutton that's\n\
# currently posted.  It takes no arguments.  Special notes:\n\
# 1. It's important to unpost the menu before releasing the grab, so\n\
#    that any Enter-Leave events (e.g. from menu back to main\n\
#    application) have mode NotifyGrab.\n\
# 2. Be sure to enclose various groups of commands in \"catch\" so that\n\
#    the procedure will complete even if the menubutton or the menu\n\
#    or the grab window has been deleted.\n\
\n\
proc tk_mbUnpost {} {\n\
    global tk_priv\n\
    set w $tk_priv(posted)\n\
    if {$w != \"\"} {\n\
	catch {\n\
	    set menu [lindex [$w config -menu] 4]\n\
	    $menu unpost\n\
	    $menu config -activebackground $tk_priv(activeBg)\n\
	    $menu config -activeforeground $tk_priv(activeFg)\n\
	    $w config -relief $tk_priv(relief)\n\
	}\n\
	catch {$tk_priv(grab) config -cursor $tk_priv(cursor)}\n\
	catch {focus $tk_priv(focus)}\n\
	grab release $tk_priv(grab)\n\
	set tk_priv(grab) \"\"\n\
	set tk_priv(focus) \"\"\n\
	set tk_priv(posted) {}\n\
    }\n\
}\n\
\n\
# The procedure below is invoked to implement keyboard traversal to\n\
# a menu button.  It takes two arguments:  the name of a window where\n\
# a keystroke originated, and the ascii character that was typed.\n\
# This procedure finds a menu bar by looking upward for a top-level\n\
# window, then looking for a window underneath that named \"menu\".\n\
# Then it searches through all the subwindows of \"menu\" for a menubutton\n\
# with an underlined character matching char.  If one is found, it\n\
# posts that menu.\n\
\n\
proc tk_traverseToMenu {w char} {\n\
    global tk_priv\n\
    if {$char == \"\"} {\n\
	return\n\
    }\n\
    set char [string tolower $char]\n\
\n\
    foreach mb [tk_getMenuButtons $w] {\n\
	if {[winfo class $mb] == \"Menubutton\"} {\n\
	    set char2 [string index [lindex [$mb config -text] 4] \\\n\
		    [lindex [$mb config -underline] 4]]\n\
	    if {[string compare $char [string tolower $char2]] == 0} {\n\
		tk_mbPost $mb\n\
		[lindex [$mb config -menu] 4] activate 0\n\
		return\n\
	    }\n\
	}\n\
    }\n\
}\n\
\n\
# The procedure below is used to implement keyboard traversal within\n\
# the posted menu.  It takes two arguments:  the name of the menu to\n\
# be traversed within, and an ASCII character.  It searches for an\n\
# entry in the menu that has that character underlined.  If such an\n\
# entry is found, it is invoked and the menu is unposted.\n\
\n\
proc tk_traverseWithinMenu {w char} {\n\
    if {$char == \"\"} {\n\
	return\n\
    }\n\
    set char [string tolower $char]\n\
    set last [$w index last]\n\
    if {$last == \"none\"} {\n\
	return\n\
    }\n\
    for {set i 0} {$i <= $last} {incr i} {\n\
	if [catch {set char2 [string index \\\n\
		[lindex [$w entryconfig $i -label] 4] \\\n\
		[lindex [$w entryconfig $i -underline] 4]]}] {\n\
	    continue\n\
	}\n\
	if {[string compare $char [string tolower $char2]] == 0} {\n\
	    tk_mbUnpost\n\
	    $w invoke $i\n\
	    return\n\
	}\n\
    }\n\
}\n\
\n\
# The procedure below takes a single argument, which is the name of\n\
# a window.  It returns a list containing path names for all of the\n\
# menu buttons associated with that window's top-level window, or an\n\
# empty list if there are none.\n\
\n\
proc tk_getMenuButtons w {\n\
    global tk_priv\n\
    set top [winfo toplevel $w]\n\
    if [catch {set bar [set tk_priv(menuBarFor$top)]}] {\n\
	return \"\"\n\
    }\n\
    return $tk_priv(menusFor$bar)\n\
}\n\
\n\
# The procedure below is used to traverse to the next or previous\n\
# menu in a menu bar.  It takes one argument, which is a count of\n\
# how many menu buttons forward or backward (if negative) to move.\n\
# If there is no posted menu then this procedure has no effect.\n\
\n\
proc tk_nextMenu count {\n\
    global tk_priv\n\
    if {$tk_priv(posted) == \"\"} {\n\
	return\n\
    }\n\
    set buttons [tk_getMenuButtons $tk_priv(posted)]\n\
    set length [llength $buttons]\n\
    for {set i 0} 1 {incr i} {\n\
	if {$i >= $length} {\n\
	    return\n\
	}\n\
	if {[lindex $buttons $i] == $tk_priv(posted)} {\n\
	    break\n\
	}\n\
    }\n\
    incr i $count\n\
    while 1 {\n\
	while {$i < 0} {\n\
	    incr i $length\n\
	}\n\
	while {$i >= $length} {\n\
	    incr i -$length\n\
	}\n\
	set mb [lindex $buttons $i]\n\
	if {[lindex [$mb configure -state] 4] != \"disabled\"} {\n\
	    break\n\
	}\n\
	incr i $count\n\
    }\n\
    tk_mbUnpost\n\
    tk_mbPost $mb\n\
    [lindex [$mb config -menu] 4] activate 0\n\
}\n\
\n\
# The procedure below is used to traverse to the next or previous entry\n\
# in the posted menu.  It takes one argument, which is 1 to go to the\n\
# next entry or -1 to go to the previous entry.  Disabled entries are\n\
# skipped in this process.\n\
\n\
proc tk_nextMenuEntry count {\n\
    global tk_priv\n\
    if {$tk_priv(posted) == \"\"} {\n\
	return\n\
    }\n\
    set menu [lindex [$tk_priv(posted) config -menu] 4]\n\
    if {[$menu index last] == \"none\"} {\n\
	return\n\
    }\n\
    set length [expr [$menu index last]+1]\n\
    set i [$menu index active]\n\
    if {$i == \"none\"} {\n\
	set i 0\n\
    } else {\n\
	incr i $count\n\
    }\n\
    while 1 {\n\
	while {$i < 0} {\n\
	    incr i $length\n\
	}\n\
	while {$i >= $length} {\n\
	    incr i -$length\n\
	}\n\
	if {[catch {$menu entryconfigure $i -state} state] == 0} {\n\
	    if {[lindex $state 4] != \"disabled\"} {\n\
		break\n\
	    }\n\
	}\n\
	incr i $count\n\
    }\n\
    $menu activate $i\n\
}\n\
\n\
# The procedure below invokes the active entry in the posted menu,\n\
# if there is one.  Otherwise it does nothing.\n\
\n\
proc tk_invokeMenu {menu} {\n\
    set i [$menu index active]\n\
    if {$i != \"none\"} {\n\
	tk_mbUnpost\n\
	update idletasks\n\
	$menu invoke $i\n\
    }\n\
}\n\
\n\
# The procedure below is invoked to keyboard-traverse to the first\n\
# menu for a given source window.  The source window is passed as\n\
# parameter.\n\
\n\
proc tk_firstMenu w {\n\
    set mb [lindex [tk_getMenuButtons $w] 0]\n\
    if {$mb != \"\"} {\n\
	tk_mbPost $mb\n\
	[lindex [$mb config -menu] 4] activate 0\n\
    }\n\
}\n\
\n\
# The procedure below is invoked when a button-1-down event is\n\
# received by a menu button.  If the mouse is in the menu button\n\
# then it posts the button's menu.  If the mouse isn't in the\n\
# button's menu, then it deactivates any active entry in the menu.\n\
# Remember, event-sharing can cause this procedure to be invoked\n\
# for two different menu buttons on the same event.\n\
\n\
proc tk_mbButtonDown w {\n\
    global tk_priv\n\
    if {[lindex [$w config -state] 4] == \"disabled\"} {\n\
	return\n\
    }\n\
    if {$tk_priv(inMenuButton) == $w} {\n\
	tk_mbPost $w\n\
    }\n\
}\n\
# text.tcl --\n\
#\n\
# This file contains Tcl procedures used to manage Tk entries.\n\
#\n\
# $Header: /user6/ouster/wish/library/RCS/text.tcl,v 1.4 93/10/23 16:21:12 ouster Exp $ SPRITE (Berkeley)\n\
#\n\
# Copyright (c) 1992-1993 The Regents of the University of California.\n\
# All rights reserved.\n\
#\n\
# Permission is hereby granted, without written agreement and without\n\
# license or royalty fees, to use, copy, modify, and distribute this\n\
# software and its documentation for any purpose, provided that the\n\
# above copyright notice and the following two paragraphs appear in\n\
# all copies of this software.\n\
#\n\
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR\n\
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT\n\
# OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF\n\
# CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n\
#\n\
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,\n\
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY\n\
# AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS\n\
# ON AN \"AS IS\" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO\n\
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.\n\
#\n\
\n\
# The procedure below is invoked when dragging one end of the selection.\n\
# The arguments are the text window name and the index of the character\n\
# that is to be the new end of the selection.\n\
\n\
proc tk_textSelectTo {w index} {\n\
    global tk_priv\n\
\n\
    if [catch {$w index anchor}] {\n\
	$w mark set anchor $index\n\
    }\n\
    case $tk_priv(selectMode) {\n\
	char {\n\
	    if [$w compare $index < anchor] {\n\
		set first $index\n\
		set last anchor\n\
	    } else {\n\
		set first anchor\n\
		set last [$w index $index+1c]\n\
	    }\n\
	}\n\
	word {\n\
	    if [$w compare $index < anchor] {\n\
		set first [$w index \"$index wordstart\"]\n\
		set last [$w index \"anchor wordend\"]\n\
	    } else {\n\
		set first [$w index \"anchor wordstart\"]\n\
		set last [$w index \"$index wordend\"]\n\
	    }\n\
	}\n\
	line {\n\
	    if [$w compare $index < anchor] {\n\
		set first [$w index \"$index linestart\"]\n\
		set last [$w index \"anchor lineend + 1c\"]\n\
	    } else {\n\
		set first [$w index \"anchor linestart\"]\n\
		set last [$w index \"$index lineend + 1c\"]\n\
	    }\n\
	}\n\
    }\n\
    $w tag remove sel 0.0 $first\n\
    $w tag add sel $first $last\n\
    $w tag remove sel $last end\n\
}\n\
\n\
# The procedure below is invoked to backspace over one character in\n\
# a text widget.  The name of the widget is passed as argument.\n\
\n\
proc tk_textBackspace w {\n\
    $w delete insert-1c insert\n\
}\n\
\n\
# The procedure below compares three indices, a, b, and c.  Index b must\n\
# be less than c.  The procedure returns 1 if a is closer to b than to c,\n\
# and 0 otherwise.  The \"w\" argument is the name of the text widget in\n\
# which to do the comparison.\n\
\n\
proc tk_textIndexCloser {w a b c} {\n\
    set a [$w index $a]\n\
    set b [$w index $b]\n\
    set c [$w index $c]\n\
    if [$w compare $a <= $b] {\n\
	return 1\n\
    }\n\
    if [$w compare $a >= $c] {\n\
	return 0\n\
    }\n\
    scan $a \"%d.%d\" lineA chA\n\
    scan $b \"%d.%d\" lineB chB\n\
    scan $c \"%d.%d\" lineC chC\n\
    if {$chC == 0} {\n\
	incr lineC -1\n\
	set chC [string length [$w get $lineC.0 $lineC.end]]\n\
    }\n\
    if {$lineB != $lineC} {\n\
	return [expr {($lineA-$lineB) < ($lineC-$lineA)}]\n\
    }\n\
    return [expr {($chA-$chB) < ($chC-$chA)}]\n\
}\n\
\n\
# The procedure below is called to reset the selection anchor to\n\
# whichever end is FARTHEST from the index argument.\n\
\n\
proc tk_textResetAnchor {w index} {\n\
    global tk_priv\n\
    if {[$w tag ranges sel] == \"\"} {\n\
	set tk_priv(selectMode) char\n\
	$w mark set anchor $index\n\
	return\n\
    }\n\
    if [tk_textIndexCloser $w $index sel.first sel.last] {\n\
	if {$tk_priv(selectMode) == \"char\"} {\n\
	    $w mark set anchor sel.last\n\
	} else {\n\
	    $w mark set anchor sel.last-1c\n\
	}\n\
    } else {\n\
	$w mark set anchor sel.first\n\
    }\n\
}\n\
# tk.tcl --\n\
#\n\
# Initialization script normally executed in the interpreter for each\n\
# Tk-based application.  Arranges class bindings for widgets.\n\
#\n\
# $Header: /user6/ouster/wish/library/RCS/tk.tcl,v 1.37 93/10/31 16:39:17 ouster Exp $ SPRITE (Berkeley)\n\
#\n\
# Copyright (c) 1992-1993 The Regents of the University of California.\n\
# All rights reserved.\n\
#\n\
# Permission is hereby granted, without written agreement and without\n\
# license or royalty fees, to use, copy, modify, and distribute this\n\
# software and its documentation for any purpose, provided that the\n\
# above copyright notice and the following two paragraphs appear in\n\
# all copies of this software.\n\
#\n\
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR\n\
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT\n\
# OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF\n\
# CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n\
#\n\
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,\n\
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY\n\
# AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS\n\
# ON AN \"AS IS\" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO\n\
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.\n\
\n\
# Insist on running with compatible versions of Tcl and Tk.\n\
\n\
scan [info tclversion] \"%d.%d\" a b\n\
if {$a != 7} {\n\
    error \"wrong version of Tcl loaded ([info tclversion]): need 7.x\"\n\
}\n\
scan $tk_version \"%d.%d\" a b\n\
if {($a != 3) || ($b < 4)} {\n\
    error \"wrong version of Tk loaded ($tk_version): need 3.4 or later\"\n\
}\n\
unset a b\n\
\n\
# Add Tk's directory to the end of the auto-load search path:\n\
\n\
lappend auto_path $tk_library\n\
\n\
# Turn off strict Motif look and feel as a default.\n\
\n\
set tk_strictMotif 0\n\
\n\
# ----------------------------------------------------------------------\n\
# Class bindings for various flavors of button widgets.  $tk_priv(window)\n\
# keeps track of the button containing the mouse $tk_priv(relief) saves\n\
# the original relief of the button so it can be restored when the mouse\n\
# button is released, and $tk_priv(buttonWindow) keeps track of the\n\
# window in which the mouse button was pressed.\n\
# ----------------------------------------------------------------------\n\
\n\
bind Button <Any-Enter> {tk_butEnter %W}\n\
bind Button <Any-Leave> {tk_butLeave %W}\n\
bind Button <1> {tk_butDown %W}\n\
bind Button <ButtonRelease-1> {tk_butUp %W}\n\
\n\
bind Checkbutton <Any-Enter> {tk_butEnter %W}\n\
bind Checkbutton <Any-Leave> {tk_butLeave %W}\n\
bind Checkbutton <1> {tk_butDown %W}\n\
bind Checkbutton <ButtonRelease-1> {tk_butUp %W}\n\
\n\
bind Radiobutton <Any-Enter> {tk_butEnter %W}\n\
bind Radiobutton <Any-Leave> {tk_butLeave %W}\n\
bind Radiobutton <1> {tk_butDown %W}\n\
bind Radiobutton <ButtonRelease-1> {tk_butUp %W}\n\
\n\
# ----------------------------------------------------------------------\n\
# Class bindings for entry widgets.\n\
# ----------------------------------------------------------------------\n\
\n\
bind Entry <1> {\n\
    %W icursor @%x\n\
    %W select from @%x\n\
    if {[lindex [%W config -state] 4] == \"normal\"} {focus %W}\n\
}\n\
bind Entry <B1-Motion> {%W select to @%x}\n\
bind Entry <Shift-1> {%W select adjust @%x}\n\
bind Entry <Shift-B1-Motion> {%W select to @%x}\n\
bind Entry <2> {%W scan mark %x}\n\
bind Entry <B2-Motion> {%W scan dragto %x}\n\
bind Entry <Any-KeyPress> {\n\
    if {\"%A\" != \"\"} {\n\
	%W insert insert %A\n\
	tk_entrySeeCaret %W\n\
    }\n\
}\n\
bind Entry <Delete> {tk_entryBackspace %W; tk_entrySeeCaret %W}\n\
bind Entry <BackSpace> {tk_entryBackspace %W; tk_entrySeeCaret %W}\n\
bind Entry <Control-h> {tk_entryBackspace %W; tk_entrySeeCaret %W}\n\
bind Entry <Control-d> {%W delete sel.first sel.last; tk_entrySeeCaret %W}\n\
bind Entry <Control-u> {%W delete 0 end}\n\
bind Entry <Control-v> {%W insert insert [selection get]; tk_entrySeeCaret %W}\n\
bind Entry <Control-w> {tk_entryBackword %W; tk_entrySeeCaret %W}\n\
tk_bindForTraversal Entry\n\
\n\
# ----------------------------------------------------------------------\n\
# Class bindings for listbox widgets.\n\
# ----------------------------------------------------------------------\n\
\n\
bind Listbox <1> {%W select from [%W nearest %y]}\n\
bind Listbox <B1-Motion> {%W select to [%W nearest %y]}\n\
bind Listbox <Shift-1> {%W select adjust [%W nearest %y]}\n\
bind Listbox <Shift-B1-Motion> {%W select to [%W nearest %y]}\n\
bind Listbox <2> {%W scan mark %x %y}\n\
bind Listbox <B2-Motion> {%W scan dragto %x %y}\n\
\n\
# ----------------------------------------------------------------------\n\
# Class bindings for scrollbar widgets.  When strict Motif is requested,\n\
# the bindings use $tk_priv(buttons) and $tk_priv(activeFg) to set the\n\
# -activeforeground color to -foreground when the mouse is in the window\n\
# and restore it when the mouse leaves.\n\
# ----------------------------------------------------------------------\n\
\n\
bind Scrollbar <Any-Enter> {\n\
    if $tk_strictMotif {\n\
	set tk_priv(activeFg) [lindex [%W config -activeforeground] 4]\n\
	%W config -activeforeground [lindex [%W config -foreground] 4]\n\
    }\n\
}\n\
bind Scrollbar <Any-Leave> {\n\
    if {$tk_strictMotif && ($tk_priv(buttons) == 0)} {\n\
	%W config -activeforeground $tk_priv(activeFg)\n\
    }\n\
}\n\
bind Scrollbar <Any-ButtonPress> {incr tk_priv(buttons)}\n\
bind Scrollbar <Any-ButtonRelease> {incr tk_priv(buttons) -1}\n\
\n\
# ----------------------------------------------------------------------\n\
# Class bindings for scale widgets.  When strict Motif is requested,\n\
# the bindings use $tk_priv(buttons) and $tk_priv(activeFg) to set the\n\
# -activeforeground color to -foreground when the mouse is in the window\n\
# and restore it when the mouse leaves.\n\
# ----------------------------------------------------------------------\n\
\n\
bind Scale <Any-Enter> {\n\
    if $tk_strictMotif {\n\
	set tk_priv(activeFg) [lindex [%W config -activeforeground] 4]\n\
	%W config -activeforeground [lindex [%W config -sliderforeground] 4]\n\
    }\n\
}\n\
bind Scale <Any-Leave> {\n\
    if {$tk_strictMotif && ($tk_priv(buttons) == 0)} {\n\
	%W config -activeforeground $tk_priv(activeFg)\n\
    }\n\
}\n\
bind Scale <Any-ButtonPress> {incr tk_priv(buttons)}\n\
bind Scale <Any-ButtonRelease> {incr tk_priv(buttons) -1}\n\
\n\
# ----------------------------------------------------------------------\n\
# Class bindings for menubutton widgets.  Variables used:\n\
# $tk_priv(posted) -		keeps track of the menubutton whose menu is\n\
#				currently posted (or empty string, if none).\n\
# $tk_priv(inMenuButton)-	if non-null, identifies menu button\n\
#				containing mouse pointer.\n\
# $tk_priv(relief) -		keeps track of original relief of posted\n\
#				menu button, so it can be restored later.\n\
# $tk_priv(dragging) -		if non-null, identifies menu button whose\n\
#				menu is currently being dragged in a tear-off\n\
#				operation.\n\
# $tk_priv(focus) -		records old focus window so focus can be\n\
#				returned there after keyboard traversal\n\
#				to menu.\n\
# ----------------------------------------------------------------------\n\
\n\
bind Menubutton <Any-Enter> {\n\
    set tk_priv(inMenuButton) %W\n\
    if {[lindex [%W config -state] 4] != \"disabled\"} {\n\
	if {!$tk_strictMotif} {\n\
	    %W config -state active\n\
	}\n\
    }\n\
}\n\
bind Menubutton <Any-Leave> {\n\
    set tk_priv(inMenuButton) {}\n\
    if {[lindex [%W config -state] 4] == \"active\"} {\n\
	%W config -state normal\n\
    }\n\
}\n\
bind Menubutton <1> {tk_mbButtonDown %W}\n\
bind Menubutton <Any-ButtonRelease-1> {\n\
    if {($tk_priv(posted) == \"%W\") && ($tk_priv(inMenuButton) == \"%W\")} {\n\
	[lindex [$tk_priv(posted) config -menu] 4] activate 0\n\
    } else {\n\
	tk_mbUnpost\n\
    }\n\
}\n\
\n\
# The binding below is trickier than it looks.  It's important to check\n\
# to see that another menu is posted in the \"if\" statement below.\n\
# The check is needed because some window managers (e.g. mwm in\n\
# click-to-focus mode) cause a button-press event to be preceded by\n\
# a B1-Enter event;  we don't want to process that B1-Enter event (if\n\
# we do, the grab may get mis-set so that the menu is non-responsive).\n\
\n\
bind Menubutton <B1-Enter> {\n\
    set tk_priv(inMenuButton) %W\n\
    if {([lindex [%W config -state] 4] != \"disabled\")\n\
	    && ($tk_priv(posted) != \"\")} {\n\
	if {!$tk_strictMotif} {\n\
	    %W config -state active\n\
	}\n\
	tk_mbPost %W\n\
    }\n\
}\n\
bind Menubutton <2> {\n\
    if {($tk_priv(posted) == \"\")\n\
	    && ([lindex [%W config -state] 4] != \"disabled\")} {\n\
	set tk_priv(dragging) %W\n\
	[lindex [$tk_priv(dragging) config -menu] 4] post %X %Y\n\
    }\n\
}\n\
bind Menubutton <B2-Motion> {\n\
    if {$tk_priv(dragging) != \"\"} {\n\
	[lindex [$tk_priv(dragging) config -menu] 4] post %X %Y\n\
    }\n\
}\n\
bind Menubutton <ButtonRelease-2> {set tk_priv(dragging) \"\"}\n\
\n\
# ----------------------------------------------------------------------\n\
# Class bindings for menu widgets.  $tk_priv(x) and $tk_priv(y) are used\n\
# to keep track of the position of the mouse cursor in the menu window\n\
# during dragging of tear-off menus.  $tk_priv(window) keeps track of\n\
# the menu containing the mouse, if any.\n\
# ----------------------------------------------------------------------\n\
\n\
bind Menu <Any-Enter> {set tk_priv(window) %W; %W activate @%y}\n\
bind Menu <Any-Leave> {set tk_priv(window) {}; %W activate none}\n\
bind Menu <Any-Motion> {\n\
    if {$tk_priv(window) == \"%W\"} {\n\
	%W activate @%y\n\
    }\n\
}\n\
bind Menu <1> {\n\
    if {$tk_priv(grab) != \"\"} {\n\
	grab $tk_priv(grab)\n\
    }\n\
}\n\
bind Menu <ButtonRelease-1> {tk_invokeMenu %W}\n\
bind Menu <2> {set tk_priv(x) %x; set tk_priv(y) %y}\n\
bind Menu <B2-Motion> {\n\
    if {$tk_priv(posted) == \"\"} {\n\
	%W post [expr %X-$tk_priv(x)] [expr %Y-$tk_priv(y)]\n\
    }\n\
}\n\
bind Menu <B2-Leave> { }\n\
bind Menu <B2-Enter> { }\n\
bind Menu <Escape> {tk_mbUnpost}\n\
bind Menu <Any-KeyPress> {tk_traverseWithinMenu %W %A}\n\
bind Menu <Left> {tk_nextMenu -1}\n\
bind Menu <Right> {tk_nextMenu 1}\n\
bind Menu <Up> {tk_nextMenuEntry -1}\n\
bind Menu <Down> {tk_nextMenuEntry 1}\n\
bind Menu <Return> {tk_invokeMenu %W}\n\
\n\
# ----------------------------------------------------------------------\n\
# Class bindings for text widgets. $tk_priv(selectMode) holds one of\n\
# \"char\", \"word\", or \"line\" to indicate which selection mode is active.\n\
# ----------------------------------------------------------------------\n\
\n\
bind Text <1> {\n\
    set tk_priv(selectMode) char\n\
    %W mark set insert @%x,%y\n\
    %W mark set anchor insert\n\
    if {[lindex [%W config -state] 4] == \"normal\"} {focus %W}\n\
}\n\
bind Text <Double-1> {\n\
    set tk_priv(selectMode) word\n\
    %W mark set insert \"@%x,%y wordstart\"\n\
    tk_textSelectTo %W insert\n\
}\n\
bind Text <Triple-1> {\n\
    set tk_priv(selectMode) line\n\
    %W mark set insert \"@%x,%y linestart\"\n\
    tk_textSelectTo %W insert\n\
}\n\
bind Text <B1-Motion> {tk_textSelectTo %W @%x,%y}\n\
bind Text <Shift-1> {\n\
    tk_textResetAnchor %W @%x,%y\n\
    tk_textSelectTo %W @%x,%y\n\
}\n\
bind Text <Shift-B1-Motion> {tk_textSelectTo %W @%x,%y}\n\
bind Text <2> {%W scan mark %y}\n\
bind Text <B2-Motion> {%W scan dragto %y}\n\
bind Text <Any-KeyPress> {\n\
    if {\"%A\" != \"\"} {\n\
	%W insert insert %A\n\
	%W yview -pickplace insert\n\
    }\n\
}\n\
bind Text <Return> {%W insert insert \\n; %W yview -pickplace insert}\n\
bind Text <BackSpace> {tk_textBackspace %W; %W yview -pickplace insert}\n\
bind Text <Delete> {tk_textBackspace %W; %W yview -pickplace insert}\n\
bind Text <Control-h> {tk_textBackspace %W; %W yview -pickplace insert}\n\
bind Text <Control-d> {%W delete sel.first sel.last}\n\
bind Text <Control-v> {\n\
    %W insert insert [selection get]\n\
    %W yview -pickplace insert\n\
}\n\
tk_bindForTraversal Text\n\
\n\
# Initialize the elements of tk_priv that require initialization.\n\
\n\
set tk_priv(buttons) 0\n\
set tk_priv(buttonWindow) {}\n\
set tk_priv(dragging) {}\n\
set tk_priv(focus) {}\n\
set tk_priv(grab) {}\n\
set tk_priv(inMenuButton) {}\n\
set tk_priv(posted) {}\n\
set tk_priv(selectMode) char\n\
set tk_priv(window) {}\n\
# This file contains a default version of the tkError procedure.  It\n\
# posts a dialog box with the error message and gives the user a chance\n\
# to see a more detailed stack trace.\n\
\n\
proc tkerror err {\n\
    global errorInfo\n\
    set info $errorInfo\n\
    if {[tk_dialog .tkerrorDialog \"Error in Tcl Script\" \\\n\
	    \"Error: $err\" error 0 OK \"See Stack Trace\"] == 0} {\n\
	return\n\
    }\n\
\n\
    set w .tkerrorTrace\n\
    catch {destroy $w}\n\
    toplevel $w -class ErrorTrace\n\
    wm minsize $w 1 1\n\
    wm title $w \"Stack Trace for Error\"\n\
    wm iconname $w \"Stack Trace\"\n\
    button $w.ok -text OK -command \"destroy $w\"\n\
    text $w.text -relief raised -bd 2 -yscrollcommand \"$w.scroll set\" \\\n\
	    -setgrid true -width 40 -height 10\n\
    scrollbar $w.scroll -relief flat -command \"$w.text yview\"\n\
    pack $w.ok -side bottom -padx 3m -pady 3m -ipadx 2m -ipady 1m\n\
    pack $w.scroll -side right -fill y\n\
    pack $w.text -side left -expand yes -fill both\n\
    $w.text insert 0.0 $info\n\
    $w.text mark set insert 0.0\n\
\n\
    # Center the window on the screen.\n\
\n\
    wm withdraw $w\n\
    update idletasks\n\
    set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \\\n\
	    - [winfo vrootx [winfo parent $w]]]\n\
    set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 \\\n\
	    - [winfo vrooty [winfo parent $w]]]\n\
    wm geom $w +$x+$y\n\
    wm deiconify $w\n\
}\n\
";
