#include "generic_prototypes.h"
#include "www_prototypes.h"
/*bus.c*/
int bus_recv();
int bus_send(char *msg, int len);
int bus_listen();
int is_a_bus_message(char *buf);
void parse_bus_message(char *buf, int len);
int bus_send_new_app();

/*random*/
#ifndef NORANDPROTO
void srandom(int seed);
int random();
#endif

/*sip.c*/
int sip_recv();
struct in_addr look_up_address(char *hostname);
int send_sip_msg(char *msg, char *dst);
int sip_listen(char *address, int port);
int sip_tx_init(char *address, int port, char ttl);
char *sip_get_dstname(char *msg);
int is_a_sip_request(char *msg);
int is_a_sip_reply(char *msg);
int parse_sip_reply(char *msg);
int parse_sip_success(char *msg);
int parse_sip_progress(char *msg);
int parse_sip_fail(char *msg);
int parse_sip_fa(char *msg);
int parse_sip_ringing(char *msg);
int parse_sip_trying(char *msg);
int sip_send(char *msg, int len, struct sockaddr_in *dst, int ttl);

/*sap_crypt.c*/
int encrypt_announcement(char *srcbuf, char **dstbuf, int *length, 
                         char *key);
int decrypt_announcement(char *buf, int *len, char *recvkey);
int load_crypted_file(char *filename, char *buf, char *key);
int write_crypted_file(char *filename, char *data, int len, char *key);
char *get_pass_phrase();
int load_keys();
int save_keys();
int get_sdr_home(char *str);
int find_key_by_name(char *keyname, u_int *keyid, char *key);
int init_security();
int set_pass_phrase(char *newphrase);
int register_key(char *key, char *keyname, u_int keyid);
int delete_key(char *keyname);

/*ui_fns.c*/
int ui_quit();
int ui_sip_send_msg(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int ui_lookup_host(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int ui_sd_listen(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int ui_generate_port();
int ui_generate_address();
int generate_id();
int ui_check_address(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int getdayname(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int getmonname(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int gettime(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);

int getemailaddress(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int uigettimeofday();
int createsession(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int getusername(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int gethostaddr(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int ui_stop_session_ad(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
void initnames();
int ui_set_passphrase(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int ui_get_passphrase(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int ui_add_key(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int ui_delete_key(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int ui_find_key_by_name(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int ui_write_crypted_file(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int ui_load_keys(ClientData dummy, Tcl_Interp *interp, int argc, char **argv);
int ui_save_keys();


/*generate_ids*/
int generate_port(char *s);
int store_address(struct in_addr *addr, unsigned long endtime);
struct in_addr generate_address();
int check_address(struct in_addr *addr);


/*byte_fns.c*/
#ifdef SYSV
void bcopy(char *s1, char *s2, int len);
void bzero(char *str, int len);
#endif

/*bitmaps.c*/
void init_bitmaps();

/*ui_init.c*/
int ui_create_interface();
int ui_init(int *argc, char **argv);
int announce_error(int code, char *command);

/*sd_listen.c*/
void rebuild_interface();
void remove_interface();
int xremove_interface(Display *d);
int sd_listen(char *address, int port, int *rxsock, int *no_of_socks, int fatal);
void recv_packets();
int timed_send_advert(ClientData cd);
int send_advert(char *adstr, int sock, unsigned char ttl, u_int keyid, 
		u_int len);
int queue_ad_for_sending(char *aid, char *adstr, int interval, long end_time, char * address, int port, unsigned char ttl, char * keyname);
int stop_session_ad(char *aid);
void clean_up_and_die();
void force_numeric(char *str);
void splat_tcl_special_chars(char *str);
void warn_tcl_special_chars(char *str);
unsigned long parse_entry(char *advertid, char *data, int length,
            unsigned long src, unsigned long hfrom,
            char *sd_addr, int port, time_t t, char *trust, char *recvkey);
void read_old_style_cache();
#ifdef LISTEN_FOR_SD
void read_sd_cache();
#endif
int extract_ttl(char *addrstr);
int check_net_type(char *in, char *ip);

/* iohandler.c */
void linksocket(int fd, int mask, Tcl_FileProc* callback);
void unlinksocket(int fd);
