/*
 * Program to listen for SD style packets and print out the
 * announcements.
 * 
 * This program is experimental and not guarenteed to digest all
 * SD packets correctly. Since the packet format was reverse
 * engineered, packets not seen before may not work.
 *
 * Tom Pusateri (pusateri@cs.duke.edu)
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted without restriction.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: sd_listen.c,v 1.6 1994/08/30 12:06:06 jon Exp jon $
 *
 * Hacked up a bit by J.P.Knight@lut.ac.uk
 *
 * $Log: sd_listen.c,v $
 * Revision 1.6  1994/08/30  12:06:06  jon
 * Now has the correct compiler preprocessor definitions for spotting a DEC
 * Alpha and adjusting the structure elements accordingly.
 *
 * Revision 1.5  1994/08/29  08:41:24  jon
 * Included support for DEC Alpha machines (and anyone else who has 64bit
 * unsigned longs).  You need to give -DALPHA on the command line to get
 * the right definitions in this case.
 *
 * Revision 1.4  1994/08/22  17:07:16  jon
 * Increased buffer size to 2K as long descriptions were screwing things up.
 * This is longer than the max Ethernet packet size so is probably ok.  If
 * dumps appear with lengths of 2048 it will be indicative of hitting this
 * limit again and will mean that the buffer size will need to be upped
 * again.  Also added extended dumping of strange packets/error conditions
 * and tweaked it so that these conditions don't mean that the program is
 * exited completely.
 *
 * Revision 1.3  1994/06/16  14:57:41  jon
 * Changed command line to make interface optional.  If not specified, it uses
 * the host's default multicast interface.
 *
 * Revision 1.2  1994/06/16  14:49:49  jon
 * Fixed segmentation fault on SunOS4.1.3 caused by the failure to printout
 * the source address from an in_addr structure rather than a simple u_long.
 * Also added in a ``fmt:unknown'' default string for when the packet does
 * not appear to have a format included to stop random rubbish being printed.
 *
 * Revision 1.1  1994/06/16  14:45:29  jon
 * Initial revision
 *
 */


#define MULTICAST

#include <locale.h>
#include <signal.h>
#include <fcntl.h>
#include "sdr.h"
#include "gateway_prototypes.h"

#define SD_GROUP        0xe0027fff
#define SD_PORT         9876
#define SDR_GROUP        0xe0027ffe
#define SDR_PORT         9875
#define MAXMEDIA 5
#define MAXVARS 256
#define MAXNAME 80

extern Tcl_Interp *interp;
unsigned long parse_sd_entry(char *session, int length,
            unsigned long src, unsigned long hfrom,
            time_t t);
int send_advert(char *adstr, int sock, unsigned char ttl);
void recv_packets();


void dump(buf, buflen)
char *buf;
int buflen;
{
        int i;
        unsigned char c;
        printf("Unexpected packet. Dumping...\n");
        printf("Buffer length: %d\n",buflen);
        for (i=0; i<buflen; i++) {
                c=buf[i];
                printf(" %02x %c", c,c);
        }
        printf("\n");
}

void seedrand()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  srand48(tv.tv_usec);
}

int rxsock;
int txsock;
unsigned long hostaddr;
char hostname[256];
char username[256];
unsigned char ttl=1;

int main(argc, argv)
int argc;
char *argv[];
{
    struct sockaddr_in name;
    struct ip_mreq imr;
    char *interface = NULL;

    struct hostent *hstent;
    struct passwd *pswd;
    uid_t uid;
    fd_set fdset;

    seedrand();

    /*find our own address*/
    gethostname(hostname, 256);
    hstent=(struct hostent*)gethostbyname(hostname);
    bcopy((char *)hstent->h_addr_list[0], (char *)&hostaddr, hstent->h_length);
    hostaddr=ntohl(hostaddr);
    
    /*find the user's username*/
    uid=getuid();
    pswd=getpwuid(uid);
    strcpy(username, pswd->pw_name);

#ifdef NOTDEF
    setlocale(LC_ALL, "");
    initnames();
    
    if (argc == 2) {
        interface=argv[1];
    }    
#endif

    if((rxsock=socket( AF_INET, SOCK_DGRAM, 0 )) < 0) {
        perror("socket");
        exit(1);
    }

    fcntl(rxsock, F_SETFD, 1);

    imr.imr_multiaddr.s_addr = SD_GROUP;
    imr.imr_multiaddr.s_addr = htonl(imr.imr_multiaddr.s_addr);
    if (interface!=NULL) {
        imr.imr_interface.s_addr = inet_addr(interface);
    } else {
        imr.imr_interface.s_addr = htonl(INADDR_ANY);
    }
    imr.imr_interface.s_addr = htonl(imr.imr_interface.s_addr);

    if (setsockopt(rxsock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
        (char *)&imr, sizeof(struct ip_mreq)) < 0 ) {
        perror("setsockopt - IP_ADD_MEMBERSHIP");
        exit(1);
    }

        /*
         *      Use INADDR_ANY if your multicast port doesn't allow
         *      binding to a multicast address.
         *
         */

    name.sin_family = AF_INET;
#ifndef CANT_MCAST_BIND
    name.sin_addr.s_addr = htonl(SD_GROUP);
#else
    name.sin_addr.s_addr = INADDR_ANY;
#endif
    name.sin_port = htons(SD_PORT);
    if (bind(rxsock, (struct sockaddr *)&name, sizeof(name))) {
        perror("bind");
        exit(1);
    }


/*Set up Tx Socket*/
    if((txsock=socket( AF_INET, SOCK_DGRAM, 0 )) < 0) {
        perror("socket");
        exit(1);
    }

    fcntl(txsock, F_SETFD, 1);

    name.sin_family = AF_INET;
    name.sin_addr.s_addr = INADDR_ANY;
    name.sin_port = 0;
#ifdef NOTDEF
    if (bind(txsock, (struct sockaddr *)&name, sizeof(name))) {
        perror("bind");
        exit(1);
    }
#endif
    if (setsockopt(txsock, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&ttl, 
		   sizeof(ttl))<0)
      {
	perror("setsockopt ttl");
	exit(-1);
      }
    name.sin_family = AF_INET;
    name.sin_addr.s_addr = htonl(SDR_GROUP);
    name.sin_port = htons(SDR_PORT);
    if (connect(txsock, (struct sockaddr *)&name, sizeof(struct sockaddr_in))<0)
      {
	perror("connect");
	fprintf(stderr, "Dest Address problem\n");
	exit(-1);
      }

#ifdef NOTDEF
    ui_init(argc, argv);
    ui_create_interface(rxsock);
#endif
    while(1)
      {
	 FD_ZERO(&fdset);
	 FD_SET(rxsock, &fdset);
	 select(rxsock+1, &fdset, 0, 0, NULL);
	 if (FD_ISSET(rxsock, &fdset)!=0)
	   {
	     recv_packets();
	   }
      }

}

void recv_packets()
{
    int length;
    struct sd_header *bp;
    static char buf[2048];
    struct sockaddr_in from;
    int fromlen;
    struct timeval tv;
    struct timezone tzp;

    fromlen=sizeof(struct sockaddr);
    if ((length = recvfrom(rxsock, (char *) buf, 2048, 0,
		       (struct sockaddr *)&from, (int *)&fromlen)) < 0) {
      perror("recv");
      exit(1);
    }
    gettimeofday(&tv, &tzp);

    /* print session */
    bp = (struct sd_header *) buf;
    parse_sd_entry(bp->name, length-8, ntohl(bp->src), ntohl(from.sin_addr.s_addr), tv.tv_sec);
}


unsigned long parse_sd_entry(char *session, int length, 
	    unsigned long src, unsigned long hfrom,
	    time_t t)
{
    int i, j;
    int origlen;
    static char namestr[2048];
    static char announce[2048];
    static char tmpline[2048];
    char *cur, *end, *fmt, *unknown, *desc, *orig, *chan, *media[MAXMEDIA];
    char *format[MAXMEDIA];
    char vars[MAXMEDIA][MAXVARS];
    char username[MAXNAME];
    char debug=0;

    char tmpstr[100], heardfrom[256], origsrc[256];
    int ttl;
    int port1, port2;
    unsigned int time1, time2;
    unsigned int aid;
    struct in_addr source;


    j = -1;
    if(strncmp(session, "s=", 2)!=0)
      {
	fprintf(stderr, "No session name field\n");
	dump(session, length);
	goto errorleap;
      }
    else
      {
	session+=2;
	length-=2;
      }
    if(((end=strchr(session, 0x0a)) == NULL)||(debug == 1)) {
      fprintf(stderr, "No end to session name\n");
      dump(session, length);
      goto errorleap;
    }
    *end++ = '\0';
#ifdef DEBUG
    printf("%s\n", session);
#endif
    origlen=length;
    length -= end-session;
    if (hfrom!=0)
      { 
	source.s_addr=htonl(hfrom);
	strncpy(heardfrom, (char *)inet_ntoa(source), 16);
      }
    if (src!=0)
      {
	source.s_addr=htonl(src);
	strncpy(origsrc, (char *)inet_ntoa(source), 16);
      }

    i = 0;
    while (length > 0) 
      {
                cur = end;
                switch (*cur) {

                case 'i':
                        /* print description */
                        desc = end+2;
                        if ((end=strchr(desc, 0x0a)) == NULL) {
                                printf("Error decoding description\n");
				printf("Failure at byte %ld\n", (long)desc-(long)session);
                                dump(session, origlen);
                                goto errorleap;
                        }
                        *end++ = '\0';
                        length -= end-cur;
			/*don't relay moved sessions*/
			if (strstr(desc, "**MOVED**")!=NULL)
			  return 0;
                        break;

                case 'o':
                        /* print originator */
                        orig = end+2;
                        if ((end=strchr(orig, 0x0a)) == NULL) {
                                printf("Error decoding originator\n");
				printf("Failure at byte %ld\n", (long)orig-(long)session);
                                dump(session, origlen);
                                goto errorleap;
                        }
                        *end++ = '\0';
                        length -= end-cur;
                        break;

                case 'c':
                        /* print channel */
                        chan = end+2;
                        if ((end=strchr(chan, 0x0a)) == NULL) {
                                printf("Error decoding channel\n");
				printf("Failure at byte %ld\n", (long)chan-(long)session);
                                dump(session, origlen);
                                goto errorleap;
                        }
                        *end++ = '\0';
                        length -= end-cur;
                        break;

                case 'm':
                        /* print media */
                        media[++j] = end+2;
                        if ((end=strchr(media[j], 0x0a)) == NULL) {
                                printf("Error decoding media\n");
				printf("Failure at byte %ld\n", (long)media[j]-(long)session);
                                dump(session, origlen);
                                goto errorleap;
                        }
			strcpy(vars[j], "");
			format[j]=NULL;
                        *end++ = '\0';
                        length -= end-cur;
                        break;
                case 'a':
                        /* print format */
                        fmt = end+2;
                        if ((end=strchr(fmt, 0x0a)) == NULL) {
                                printf("Error decoding attribute\n");
				printf("Failure at byte %ld\n", (long)fmt-(long)session);
                                dump(session, origlen);
                                goto errorleap;
                        }
                        *end++ = '\0';
                        length -= end-cur;
			if (strncmp(fmt,"fmt:",4)==0)
			  {
			    format[j]=fmt+4;
			  }
			else 
			  {
			    if (strcmp(vars[j], "")!=0) 
			      {
				strcat(vars[j], "\na=");
			      }
			    strcat(vars[j], fmt);
			  }
                        break;
		case 'n':
			/*sd cache extra data - not in packet stream!*/
			unknown = end+2;
			printf("decoding cache data!\n");
                        if ((end=strchr(unknown, 0x0a)) == NULL) {
                                printf("Error decoding cache data\n");
				printf("Failure at byte %ld\n", (long)unknown-(long)session);
                                dump(session, origlen);
                                goto errorleap;
                        }
                        *end++ = '\0';
                        length -= end-cur;
#ifdef NOTDEF
			sscanf(unknown, "%u %u %u", &u1, &u2, &u3);
			printf("%u %u %u\n", u1, u2, u3);
			cftime(namestr, "%d %b %y %H:%M %Z", &u3);
			t=u3;
			printf("%s\n", namestr);
			source.s_addr=u2;
			printf("Heard from: %s\n", inet_ntoa(htonl(source)));
			strcpy(heardfrom, (char *)inet_ntoa(htonl(source)));
			source.s_addr=u1;
			printf("Original src: %s\n", inet_ntoa(htonl(source)));
			strcpy(origsrc, (char *)inet_ntoa(htonl(source)));
#endif
			break;
                default:
                        /* unknown */
                        unknown = end+2;
                        printf("Warning: unknown option - >%s<\n", end);
                        if ((end=strchr(unknown, 0x0a)) == NULL) {
                                printf("Error decoding unknown\n");
				printf("Failure at byte %ld\n", (long)unknown-(long)session);
                                dump(session, origlen);
                                goto errorleap;
                        }
                        *end++ = '\0';
                        length -= end-cur;
                        break;
                }
                i++;
	      }
    j++;

    sprintf(namestr, "%s%s", orig, session);
    /*Create a hash of originator and session name as advert ID*/
    aid=0;
    for(i=0;i<=4;i++)
      namestr[strlen(namestr)+i]='\0';
    for(i=0;i<=strlen(namestr)/4;i++)
      {
	unsigned int tmp;
	bcopy(&namestr[i*4],(char *)&tmp,4);
#ifdef NOTDEF
	printf("%c%c%c%c-%x,",
	       namestr[i*4],
	       namestr[1+i*4],
	       namestr[2+i*4],
	       namestr[3+i*4],
	       tmp);
#endif
	aid=(aid<<1|aid>>31)^tmp;
      }

    announce[0]='\0';
    sprintf(tmpline, "v=0\n");
    strcat(announce, tmpline);
    source.s_addr=htonl(src);
    strncpy(username, orig, MAXNAME);
    if(strchr(username,'@')!=NULL)
      {
	*strchr(username,'@')='\0';
      }
    while(strchr(username,' ')!=0)
      {
	*strchr(username,' ')='_';
      }
    if(strcmp(username, "")==0)
      {
	strcpy(username, "unknown");
      }
    sprintf(tmpline, "o=%s %u %u IN IP4 %s\n", username, aid, 0, inet_ntoa(source));
    strcat(announce, tmpline);

    sprintf(tmpline, "s=%s\n", session);
    strcat(announce, tmpline);
    sprintf(tmpline, "i=%s\n", desc);
    strcat(announce, tmpline);
    sprintf(tmpline, "e=relayed from sd <%s>\n", orig);
    strcat(announce, tmpline);
    sprintf(tmpline, "p=unknown +0 0\n");
    strcat(announce, tmpline);

    sscanf(chan, "%s %d %u %u", tmpstr, &ttl, &time1, &time2);
    sprintf(tmpline, "c=IN IP4 %s/%d\n", tmpstr, ttl);
    strcat(announce, tmpline);
    if(time1!=0)
      {
        time1+=123010304;
        time2+=123010304;
      }

    sprintf(tmpline, "t=%u %u\n", time1, time2);
    strcat(announce, tmpline);

    for(i=0;i<j;i++)
      {
	sscanf(media[i], "%s %d %d", tmpstr, &port1, &port2);
	if (strcmp(tmpstr, "audio")==0)
	  {
	    char fmt[10];
	    if (format[i]==NULL)
	      strcpy(fmt, "pcm");
	    else
	      strcpy(fmt, format[i]);
	    sprintf(tmpline, "m=audio %d vat %s\n", port1, fmt);
	    strcat(announce, tmpline);
	    if (port2!=0) 
	      {
		sprintf(tmpline, "a=id:%d\n", port2);
		strcat(announce, tmpline);
	      }
	  }
	else if (strcmp(tmpstr, "video")==0)
	  {
	    char fmt[10];
	    if (format[i]==NULL)
	      strcpy(fmt, "nv");
	    else
	      strcpy(fmt, format[i]);
	    sprintf(tmpline, "m=video %d rtp %s\n", port1, fmt);
	    strcat(announce, tmpline);
	  }
	else if (strcmp(tmpstr, "whiteboard")==0)
	  {
	    char fmt[10];
	    if (format[i]==NULL)
	      strcpy(fmt, "wb");
	    else
	      strcpy(fmt, format[i]);
	    sprintf(tmpline, "m=whiteboard %d udp %s\n", port1, fmt);
	    strcat(announce, tmpline);
	  }
	if(strcmp(vars[i],"")!=0)
	  {
	    sprintf(tmpline, "a=%s\n", vars[i]);
	    strcat(announce, tmpline);
	  }
      }

    send_advert(announce, txsock, (unsigned char)ttl);

    return time2;
errorleap:
    return 0;
}


int send_advert(char *adstr, int sock, unsigned char ttl)
{
  char *buf;
  struct sd_header *bp;
  buf=(char *)malloc(sizeof(struct sd_header)+strlen(adstr));
  bp=(struct sd_header *)buf;
  strcpy(bp->name, adstr);
  bp->src=htonl(hostaddr);
  bp->unknown1=0;

/*  dump(buf, sizeof(struct sd_header)+strlen(adstr)-2);*/
    if (setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&ttl, 
		   sizeof(ttl))<0)
      {
	perror("setsockopt ttl");
	exit(-1);
      }
  send(sock, buf, sizeof(struct sd_header)+strlen(adstr)-4, 0);
  return 0;
}
