
/*
 * Copyright (c) 1995 University College London
 * Copyright (c) 1994 Tom Pusateri, J.P.Knight
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Computer Science
 *      Department at University College London
 * 4. Neither the name of the University nor of the Department may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


/*
 * Parts of this code were originally written by
 * Tom Pusateri (pusateri@cs.duke.edu)
 * J.P.Knight@lut.ac.uk
 * as part of sd_listen.c
 *
 * not that much of the original remains now...  MJH
 */


#define MULTICAST

#include <locale.h>
#include <signal.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "sdr.h"
#include "sdr_server.h"
#include "store_entries.h"
#include "server_send.h"
#include "server_prototypes.h"

#define SD_GROUP        "224.2.127.254"
#define LSDAP_GROUP     "224.2.127.253"
#define SD_PORT         9875
#define LSDAP_PORT      9898
#define MAXMEDIA 10
#define MAXPHONE 10
#define MAXBW 10
#define MAXKEY 10
#define MAXTIMES 10
#define MAXRPTS 10
#define MAXVARS 20
#define MAX_SOCKS 20
#define TIMEOUT 3

extern int httpport;
extern struct entry_list entries;
extern struct send_queue rq;

void dump(char *buf, int buflen)
{
#ifdef DEBUG
        int i;
        unsigned char c;
        printf("Unexpected packet. Dumping...\n");
        printf("Buffer length: %d\n",buflen);
        for (i=0; i<buflen; i++) {
                c=buf[i];
#ifdef HEXDEBUG
                printf(" %02x %c", c,c);
#else
		printf("%c",c);
#endif
        }
        printf("\n");
#endif
}

void seedrand()
{
  struct timeval tv;
#ifndef SYSV
  gettimeofday(&tv, NULL);
#else
  gettimeofday(&tv);
#endif
  srand48(tv.tv_usec);
}

int rxsock[MAX_SOCKS];
char *rx_sock_addr[MAX_SOCKS];
int rx_sock_port[MAX_SOCKS];
int no_of_rx_socks=0;
int txsock[MAX_SOCKS];
char *tx_sock_addr[MAX_SOCKS];
int no_of_tx_socks=0;
unsigned long hostaddr;
char hostname[256];
char *homedir;
char username[256];
unsigned char lzone_ttl=16;
int querier=1;
int time_since_query=TIMEOUT;
int use_cache=1;

int sd_listen(char *address, int port, int rx_sock[], int *no_of_socks) 
{    
    struct sockaddr_in name;
    struct ip_mreq imr;
    char *interface = NULL;
    char addr[16];
    unsigned int group, b0, b1, b2, b3;
    int i,one=1;
    printf("sd_rx: address: %s\n", address);
    for(i=0;i<*no_of_socks;i++)
	if (strcmp(address, rx_sock_addr[i])==0) {return(*no_of_socks);}
    printf("sd_rx: really setting address: %s\n", address);

    strncpy(addr,address,16);
    b0=atoi(strtok(addr,"."));
    b1=atoi(strtok((char*)NULL,"."));
    b2=atoi(strtok((char*)NULL,"."));
    b3=atoi(strtok((char*)NULL,"."));

    group = (((((b0*256)+b1)*256)+b2)*256)+b3;
    if((rx_sock[*no_of_socks]=socket( AF_INET, SOCK_DGRAM, 0 )) < 0) {
        perror("socket");
        exit(1);
    }

    fcntl(rx_sock[*no_of_socks], F_SETFD, 1);

    imr.imr_multiaddr.s_addr = htonl(group);
    if (interface!=NULL) {
        imr.imr_interface.s_addr = inet_addr(interface);
    } else {
        imr.imr_interface.s_addr = htonl(INADDR_ANY);
    }
    imr.imr_interface.s_addr = htonl(imr.imr_interface.s_addr);

    if (setsockopt(rx_sock[*no_of_socks], IPPROTO_IP, IP_ADD_MEMBERSHIP,
        (char *)&imr, sizeof(struct ip_mreq)) < 0 ) {
        perror("setsockopt - IP_ADD_MEMBERSHIP");
        exit(1);
    }

        /*
         *      Use INADDR_ANY if your multicast port doesn't allow
         *      binding to a multicast address.
         *
         */

    name.sin_family = AF_INET;
#ifndef CANT_MCAST_BIND
    name.sin_addr.s_addr = htonl(group);
#else
    name.sin_addr.s_addr = INADDR_ANY;
#endif
    name.sin_port = htons(port);
    if (bind(rx_sock[*no_of_socks], (struct sockaddr *)&name, sizeof(name))) {
        perror("bind");
        exit(1);
    }
    setsockopt(rx_sock[*no_of_socks], SOL_SOCKET, SO_REUSEADDR,
               (char *)&one, sizeof(one));
    rx_sock_addr[*no_of_socks]=malloc(strlen(address)+1);
    strcpy(rx_sock_addr[*no_of_socks], address);
    rx_sock_port[*no_of_socks]=port;
#ifdef DEBUG
    printf("sock %d bound to %x %d (%s)\n",*no_of_socks,group,port,address);
#endif
    (*no_of_socks)++;
    return(*no_of_socks);
}

int sd_tx(char *address, int port, unsigned char ttl, int *txsock, int *no_of_socks) 
{
    struct sockaddr_in name;
    char addr[16];
    unsigned int group, b0, b1, b2, b3;
    int i;
    char one=1, zero=0;
    printf("sd_tx: address: %s, ttl: %d\n", address, ttl);
    for(i=0;i<*no_of_socks;i++)
      if (strcmp(address, tx_sock_addr[i])==0) {return(txsock[i]);}
    tx_sock_addr[*no_of_socks]=malloc(strlen(address)+1);
    strcpy(tx_sock_addr[*no_of_socks],address);
    printf("sd_tx: really setting address: %s, ttl: %d\n", address, ttl);

    
    strncpy(addr,address,16);
 
    b0=atoi(strtok(addr,"."));
    b1=atoi(strtok((char*)NULL,"."));
    b2=atoi(strtok((char*)NULL,"."));
    b3=atoi(strtok((char*)NULL,"."));

    group = (((((b0*256)+b1)*256)+b2)*256)+b3;
    if((txsock[*no_of_socks]=socket( AF_INET, SOCK_DGRAM, 0 )) < 0) {
        perror("socket");
        exit(1);
    }

    fcntl(txsock[*no_of_socks], F_SETFD, 1);

    name.sin_family = AF_INET;
    name.sin_addr.s_addr = INADDR_ANY;
    name.sin_port = 0;
#ifdef NOTDEF
    if (bind(txsock[*no_of_socks], (struct sockaddr *)&name, sizeof(name))) {
        perror("bind");
        exit(1);
    }
#endif
    if (setsockopt(txsock[*no_of_socks], IPPROTO_IP, IP_MULTICAST_TTL, (char *)&ttl, 
		   sizeof(ttl))<0)
      {
	
	fprintf(stderr, "ttl: %d\n", ttl);
	perror("setsockopt ttl");
	exit(-1);
      }
    name.sin_family = AF_INET;
    name.sin_addr.s_addr = htonl(group);
    name.sin_port = htons(port);
    if (connect(txsock[*no_of_socks], (struct sockaddr *)&name, sizeof(struct sockaddr_in))<0)
      {
	perror("connect");
	fprintf(stderr, "Dest Address problem\n");
	exit(-1);
      }
    setsockopt(txsock[*no_of_socks], SOL_SOCKET, SO_REUSEADDR,
               (char *)&one, sizeof(one));
    if (setsockopt(txsock[*no_of_socks], IPPROTO_IP, IP_MULTICAST_LOOP, &zero, 1) < 0) 
      {
	perror("IP_MULTICAST_LOOP");
	exit(1);
    }
    (*no_of_socks)++;
    return(txsock[*no_of_socks-1]);
}


int main(argc, argv)
int argc;
char *argv[];
{
    int i;
    char buf[MAXADSIZE];
    char cachename[256];
    FILE *cache;
    DIR *dirp;
    struct dirent *direntp;
    char configname[256];

    struct hostent *hstent;
    struct passwd *pswd;
    uid_t uid;
    struct timeval tv;
    int http_sock;

    seedrand();
    signal(SIGINT, (void(*))clean_up_and_die);
    signal(SIGQUIT, (void(*))clean_up_and_die);
    signal(SIGTERM, (void(*))clean_up_and_die);

    /*find our own address*/
    gethostname(hostname, 256);
    hstent=(struct hostent*)gethostbyname(hostname);
    bcopy((char *)hstent->h_addr_list[0], (char *)&hostaddr, hstent->h_length);
    hostaddr=ntohl(hostaddr);
    
    /*find the user's username*/
    uid=getuid();
    pswd=getpwuid(uid);
    if(pswd!=0)
      strcpy(username, pswd->pw_name);
    else
      strcpy(username, "noname");

#ifdef NOTDEF
    setlocale(LC_ALL, "");
    initnames();
#endif
    
    lzone_ttl=(unsigned char)16;
    strcpy(configname, "sdr.conf");
    for(i=1;i<argc;i++)
      {
	if((strncmp(argv[i], "-t", 3)==0) && ((i+1)<argc))
	  {
/*	    if (strlen(argv[i+1]) > 3) goto error;*/
	    lzone_ttl=atoi(argv[i+1]);
	    argv[i]="";
	    argv[i+1]="";
	  }
	if((strncmp(argv[i], "-p", 3)==0) && ((i+1)<argc))
	  {
	    httpport=atoi(argv[i+1]);
	    argv[i]="";
	    argv[i+1]="";
	  }
	if((strncmp(argv[i], "-c", 3)==0) && ((i+1)<argc))
	  {
	    strcpy(configname, argv[i+1]);
	    argv[i]="";
	    argv[i+1]="";
	  }
	if(strncmp(argv[i], "-nocache", 4)==0)
	  {
	    use_cache=0;
	  }
      }

    no_of_rx_socks=0;
    no_of_tx_socks=0;
/*Set up Tx Socket*/
/*note this one must be done first*/
    sd_tx(LSDAP_GROUP, LSDAP_PORT, lzone_ttl, txsock, &no_of_tx_socks);
    if (setsockopt(txsock[0], IPPROTO_IP, IP_MULTICAST_TTL, (char *)&lzone_ttl, 
		   sizeof(lzone_ttl))<0)
      {
	fprintf(stderr, "ttl: %d\n", lzone_ttl);
	perror("setsockopt ttl");
	exit(-1);
      }

/*Set up Initial Rx Socket for local Requests*/
/*note this one must be done first*/
    sd_listen(LSDAP_GROUP, LSDAP_PORT,  rxsock, &no_of_rx_socks);

/*set up all the admin scoped sockets*/
    server_init(configname);

/*Set up send and receive sockets for Local Group and Remote Groups*/

/*Set up Initial Rx Socket*/
    sd_listen(SD_GROUP, SD_PORT, rxsock, &no_of_rx_socks);

/*Set up Tx Socket*/
/*this is the default if no admin scope config entries are given*/
/*ttl should be set when the socket is actually used*/
    sd_tx(SD_GROUP, SD_PORT, 1, txsock, &no_of_tx_socks);


#ifndef SYSV
    gettimeofday(&tv, NULL);
#else
    gettimeofday(&tv);
#endif

    /*initialise list of stored sessions*/
    init_entry_list(&entries);

    /*initialise queue of sessions to send*/
    init_send_queue(&rq);

    homedir=(char *)getenv("HOME");
    if(use_cache==1)
      {
	sprintf(cachename, "%s/.sdr_server", homedir);
	dirp = opendir( cachename );
	if (dirp==NULL)
	  {
	    if (mkdir(cachename, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH)!=0)
	      {
		fprintf(stderr, "Sdr server: Can't create cache directory\n");
	      }
	  }
	else 
	  {
	    while ( (direntp = readdir( dirp )) != NULL )
	      {
		char sd_addr[20];
		int sd_port;
		unsigned long  origsrc, src, endtime;
		time_t t;
		int ttl;
		unsigned int origin;
		struct scope *scope;
		struct stat sbuf;
		sprintf(cachename, "%s/.sdr_server/%s", homedir, direntp->d_name);
		stat(cachename, &sbuf);
		if ((sbuf.st_mode&S_IFREG)==0) continue;
		cache=fopen(cachename, "r");
		if (cache!=NULL) 
		  {
		    fgets(buf, MAXADSIZE, cache);
		    if(strncmp(buf, "n=", 2)==0)
		      {
			struct session *sess;
			sscanf(&buf[2], "%lu %lu %lu %s %u %u %u", &origsrc, 
			       &src, &t, sd_addr, &sd_port, &ttl, &origin);
			scope=find_scope(sd_addr, ttl);
			bzero(buf, MAXADSIZE);
			fread(buf, MAXADSIZE, 1, cache);
			printf("Got a cache entry\n");
			sess=parse_session(buf);
			printf("endtime: %lu, now: %lu\n", endtime, tv.tv_sec);
			if((sess->endtime>tv.tv_sec)||(sess->endtime==0))
			  {
			    printf("Storing %s\n", buf);
			    store_entry(sess, buf, strlen(buf), &entries, origin, scope);
			  }
		      }
		    else
		      {
			fprintf(stderr, "sdr:corrupted cache file: %s\n",
				cachename);
		      }
		    fclose(cache);
		  }
	      }
	    (void)closedir( dirp );
	  }
      }

    /*initialise for incoming HTTP connections*/
    http_sock=http_init();

    /*initialise HTML mapping table*/
    inithtmlmap();

    while(1) 
      {
	fd_set fdset;
	int rxsockmax;
	struct timeval tv;
	int packetflag;
	long timeout;
	tv.tv_usec=0;
	timeout=time_to_send(&rq);
	while(timeout<=0)
	  {
	    send_next_ad(&rq);
	    timeout=time_to_send(&rq);
	  }
	if(timeout<5)
	  {
	    tv.tv_sec=timeout;
	    packetflag=1;
	  }
	else
	  {
	    tv.tv_sec=5;
	    packetflag=0;
	  }
	FD_ZERO (&fdset);
	rxsockmax=0;
	for(i=0;i<no_of_rx_socks;i++)
	  {
	    FD_SET (rxsock[i], &fdset);
	    if (rxsock[i]>rxsockmax) rxsockmax=rxsock[i];
	  }
	FD_SET (http_sock, &fdset);
	http_fd_set(&fdset);
	select(FD_SETSIZE, &fdset, NULL, NULL, &tv);
	if ( FD_ISSET (rxsock[0], &fdset))
	  {
	    packetflag=1;
	    recv_local_packets();
	  }
	if ( FD_ISSET (http_sock, &fdset))
	  {
	    http_accept();
	  }
	http_fd_isset(&fdset);
	for(i=1;i<no_of_rx_socks;i++)
	  {
	    if ( FD_ISSET (rxsock[i],&fdset))
	      {
		packetflag=1;
		recv_packets(i);
	      }
	  }
	if(packetflag==0)
	  {
	    /*twas a timer event*/
	    send_hash_list(txsock[0]);
	  }
      }
}

void recv_local_packets()
{
    int length;
    struct lsdap_header *bp;
    static char buf[MAXADSIZE];
    struct sockaddr_in from;
    int fromlen;
    struct timeval tv;

    fromlen=sizeof(struct sockaddr);
    
    if ((length = recvfrom(rxsock[0], (char *) buf, MAXADSIZE, 0,
		       (struct sockaddr *)&from, (int *)&fromlen)) < 0) {
      perror("recv");
      exit(1);
    }
    if (length==MAXADSIZE) {
      /*some sneaky bugger is trying to splat the stack?*/
      fprintf(stderr, "Warning: 2K local announcement truncated\n");
    }
#ifndef SYSV
    gettimeofday(&tv, NULL);
#else
    gettimeofday(&tv);
#endif

    /* print session */
    bp = (struct lsdap_header *) buf;
    switch (bp->type)
      {
      case LSDAP_ADVERT:
	parse_local_advert((struct lsdap_advert*)buf, length);
	break;
      case LSDAP_HASHLIST:
	parse_local_hashlist((struct lsdap_hashlist*)buf, length,  ntohl(from.sin_addr.s_addr));
	break;
      case LSDAP_HASHREQ:
	parse_local_hashreq((struct lsdap_hashreq*)buf, length);
	break;
      default:
	printf("Sdr server: Unknown local message type received\n");
      }
    
}

void recv_packets(int ix)
{
    int length;
    struct sd_header *bp;
    static char buf[MAXADSIZE];
    struct sockaddr_in from;
    int fromlen;
    struct timeval tv;
    struct session *sess;
    u_int origin;

    printf("recv%d\n", (int)ix);

    fromlen=sizeof(struct sockaddr);
    
    if ((length = recvfrom(rxsock[ix], (char *) buf, MAXADSIZE, 0,
		       (struct sockaddr *)&from, (int *)&fromlen)) < 0) {
      perror("recv");
      exit(1);
    }
    if (length==MAXADSIZE) {
      /*some sneaky bugger is trying to splat the stack?*/
      fprintf(stderr, "Warning: 2K announcement truncated\n");
    }
#ifndef SYSV
    gettimeofday(&tv, NULL);
#else
    gettimeofday(&tv);
#endif
    
    bp = (struct sd_header *) buf;
    sess=parse_session(bp->name);
    printf("aid: %u, endtime: %u, now: %lu\n", sess->aid, sess->endtime, tv.tv_sec);
    if((sess->endtime>tv.tv_sec)||(sess->endtime==0))
      {
	store_entry(sess, bp->name, length-8, &entries, REMOTE, NULL);
	origin=get_origin(sess->aid, &entries);
	if(querier==1)
	  send_local_advert(sess->aid, origin, bp->name, txsock[0], length-8);
      }
}



int send_advert(char *adstr, int tx_sock, unsigned char ttl)
{
  char *buf;
  struct sd_header *bp;
  buf=(char *)malloc(sizeof(struct sd_header)+strlen(adstr));
  bp=(struct sd_header *)buf;
  strcpy(bp->name, adstr);
  bp->src=htonl(hostaddr);
  bp->unknown1=0;
    if (setsockopt(tx_sock, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&ttl, 
		   sizeof(ttl))<0)
      {
	fprintf(stderr, "ttl: %d\n", ttl);
	perror("setsockopt ttl");
	exit(-1);
      }
#ifdef DEBUG
  printf("sending ad to sock %d, ttl %d\n", tx_sock, ttl);
#endif
  send(tx_sock, buf, sizeof(struct sd_header)+strlen(adstr)-4, 0);
  return 0;
}



void clean_up_and_die()
{
  int i;
  printf("sdr_server exiting\n");
  for(i=0;i<no_of_rx_socks;i++)
    close(rxsock[i]);
  for(i=0;i<no_of_tx_socks;i++)
    close(txsock[i]);
  save_cache();
  http_closedown();
  exit(0);
}

void save_cache()
{
  char cachename[256];
  DIR *dirp;
  FILE *file;
  struct dirent *direntp;
  char *stored_buf;
  hash_t *hashlist;
  int length, newlen, i;
  if(use_cache==1)
    {
        sprintf(cachename, "%s/.sdr_server", homedir);
        dirp = opendir( cachename );
        if (dirp==NULL)
          {
            if (mkdir(cachename, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH)!=0)
              {
                fprintf(stderr, "Sdr server: Can't create cache directory\n");
              }
          }
        else
          {
            /*delete the old cache files*/
            while ( (direntp = readdir( dirp )) != NULL )
              {
                struct stat sbuf;
                sprintf(cachename, "%s/.sdr_server/%s", homedir, direntp->d_name);
                stat(cachename, &sbuf);
                if ((sbuf.st_mode&S_IFREG)==0) continue;
                unlink(cachename);
              }
            (void)closedir( dirp );
          }
 
      }
  hashlist=create_hash_list(&length, &entries);
  for(i=0;i<length;i++)
    {
      unsigned int origin;
      stored_buf=get_stored_entry(hashlist[i], &entries, &newlen);
      origin=get_origin(hashlist[i], &entries);
      sprintf(cachename, "%s/.sdr_server/%x", homedir, hashlist[i]);
      file=fopen(cachename, "w");
      fprintf(file, "n=0 0 0 224.2.127.253 9874 1 %u\n%s", origin, stored_buf);
      fclose(file);
    }
  free_hash_list(hashlist);
}

void force_numeric(char *str)
{
  /*security - splat any none numeric characters before passing this to Tcl*/
  int i;
  for(i=0;i<strlen(str);i++)
    {
      if ((str[i]<48)||(str[i]>57)) {
	str[i]=' ';
      }
    }
}

void splat_tcl_special_chars(char *str) 
{
  /*security - don't want someone passing Tcl commands in an announcement*/
  /*splat any special characters Tcl might use for execution, etc*/
  int i;
  for(i=0;i<strlen(str);i++)
    {
      if ((str[i]=='[')||(str[i]==']')||(str[i]=='$')) 
	{
	  str[i]=' ';
	}
    }
}

void warn_tcl_special_chars(char *str) 
{
  /*security - don't want someone passing Tcl commands in an announcement*/
  int i;
  for(i=0;i<strlen(str);i++)
    {
      if ((str[i]=='[')||(str[i]==']')||(str[i]=='$')) 
	{
	  fprintf(stderr, "WARNING: received key containing Tcl special chars\n");
	}
    }
}

void parse_local_advert(struct lsdap_advert *buf, int length)
{
  struct session *sess;
  char localbuf[MAXADSIZE];
  struct scope *scope;
  bzero(localbuf, MAXADSIZE);
/*  printf("Got a local advert, hash %u\n", (unsigned int)ntohl(buf->hash));*/
  bcopy(buf->addata, localbuf, length+1-sizeof(struct lsdap_advert));
  sess = parse_session(localbuf);
  scope=find_scope(sess->media->addr, sess->media->ttl);
  store_entry(sess, buf->addata, length+1-sizeof(struct lsdap_advert), &entries, buf->origin, scope);
}

void parse_local_hashlist(struct lsdap_hashlist *buf, int length, unsigned long src)
{
  int i;
  int items;
  int mylength;
/*  printf("Got a hashlist seq no %d, length %d\n", buf->seq, length);*/
  i=(length-sizeof(struct lsdap_advert));
  items=(sizeof(hash_t));
  items=(i/items)+1;
  if(querier==1)
    {
      free_hash_list(create_hash_list(&mylength, &entries));
      if(mylength<items)
	{
	  querier=0;
	  time_since_query=TIMEOUT;
	  printf("Other site has more entries\n");
	}
      else if(mylength==items)
	{
	  if(src<hostaddr)
	    {
	      printf("Other site has lower IP address\n");
	      querier=0;
	      time_since_query=TIMEOUT;
	    }
	  else
	    {
	      printf("I'm staying querier\n");
	    }
	}
      else
	{
	  printf("I have more entries\n");
	}
    }
  else
    {
      free_hash_list(create_hash_list(&mylength, &entries));
      if(mylength>items)
	{
	  printf("A\n");
	  /*I'm not the querier but the querier has less entries than me*/
	  /*could be the querier died restarted, or I missed a deletion*/
	  /*send a hash list - this should elicit a response that either
	    updates the querier or updates me*/
	  /*This should really be done after a random backoff where there
	    are more than two servers*/
	  send_hash_list(txsock[0]);
	}
      else
	{
	  printf("B\n");
	  time_since_query=TIMEOUT;
	}
    }
  for(i=0;i<items;i++)
    {
/*      printf("i:%d, hash: %u\n",i, buf->hashlist[i]);*/
      if(have_entry(buf->hashlist[i], &entries)==1)
	{
/*	  printf("have this one\n");*/
	}
      else
	{
	  printf("send_hash_request %u\n", buf->hashlist[i]);
	  send_hash_request(buf->hashlist[i], txsock[0]);
	}
    }
}

void parse_local_hashreq(struct lsdap_hashreq *buf, int length)
{
  int newlen;
  u_int origin;
  char *stored_buf;
/*  printf("Got a hashreq for hash %u\n", (unsigned int)ntohl(buf->hash));*/
  if(querier==1)
    {
      stored_buf=get_stored_entry(buf->hash, &entries, &newlen);
      origin=get_origin(buf->hash, &entries);
      if(stored_buf!=NULL)
	{
	  send_local_advert(buf->hash, origin, stored_buf, txsock[0], newlen);
	}
    }
}


void send_local_advert(u_int aid, u_int origin, char *buf, int txsock, int length)
{
  char sendbuf[MAXADSIZE];
  struct lsdap_advert *ad;
  bzero(sendbuf, MAXADSIZE);
  ad=(struct lsdap_advert *)sendbuf;
  ad->type=LSDAP_ADVERT;
  ad->check=LSDAP_CHECK_BYTE;
  ad->hash=htonl(aid);
  ad->ttl=lzone_ttl;
  ad->origin=(u_char)origin;
  bcopy(buf, ad->addata, length);
  printf("sending advert %u\n", aid);
  send(txsock, sendbuf, sizeof(struct lsdap_advert)+length-1, 0);
}

void send_hash_request(u_int aid, int txsock)
{
  struct lsdap_hashreq ad;
  ad.type=LSDAP_HASHREQ;
  ad.check=LSDAP_CHECK_BYTE;
  ad.hash=htonl(aid);
   printf("sending hash req %u\n", aid);
  send(txsock, (char *)&ad, sizeof(struct lsdap_hashreq), 0);
}

void send_hash_list(int txsock)
{
  hash_t *hashlist;
  int length;
  struct lsdap_hashlist *ad;
  char sendbuf[MAXADSIZE];
  
  if(querier==1)
    {
      bzero(sendbuf, MAXADSIZE);
      ad=(struct lsdap_hashlist *)sendbuf;
      ad->type=LSDAP_HASHLIST;
      ad->check=LSDAP_CHECK_BYTE;
      hashlist=create_hash_list(&length, &entries);
      bcopy((char *)hashlist, (char *)ad->hashlist , sizeof(hash_t)*length);
      free_hash_list(hashlist);
      printf("sending hash list\n");
      send(txsock, sendbuf, sizeof(struct lsdap_advert)+(sizeof(hash_t)*(length-1)), 0);
    }
  else 
    {
      time_since_query--;
      if(time_since_query<=0)
	querier=1;
    }
}


unsigned long ntptounix(unsigned long ntp)
{
  if(ntp==0) return(ntp);
  return(ntp-0x83aa7e80);
}

unsigned long unixtontp(unsigned long time)
{
  if(time==0) return(time);
  return(time+0x83aa7e80);
}

