#include "sdr.h"
#include "sdr_server.h"
#include "store_entries.h"
#include "server_prototypes.h"

void htmlstrcat(char *str, char* new);

extern struct entry_list entries;

int html_print_session(struct session *sess, char *str)
{
  struct media *media;
  struct lent *lent;

  strcpy(str, "<h2>");
  htmlstrcat(str, sess->sname);
  strcat(str, "</h2>");
  htmlstrcat(str, sess->desc);
  if(sess->uri!=NULL)
    {
      strcat(str, "<p><a href=");
      strcat(str, sess->uri);
      strcat(str, ">More details</a> are available.\n");
    }

  strcat(str, "<H3>Contact Details:</H3></DL>\n");
  lent=sess->email;
  while(lent!=NULL)
    {
      strcat(str, "<DT><img src=/images/email> Email: ");
      htmlstrcat(str, lent->data);
      lent=lent->next;
    }
  lent=sess->phone;
  while(lent!=NULL)
    {
      strcat(str, "<DT><img src=/images/phone> Phone: ");
      htmlstrcat(str, lent->data);
      lent=lent->next;
    }
  strcat(str, "</DL>\n");
  if(sess->endtime!=0)
    {
      struct tm *tms;
      char timestr[40];
      strcat(str, "<h3>Lifetime</h3>\nFrom ");
      tms=localtime((time_t*)&sess->starttime);
      strftime(timestr, 40, "%d %b %y %H:%M %Z", tms);
      strcat(str, timestr);
      strcat(str, " to ");
      tms=localtime((time_t*)&sess->endtime);
      strftime(timestr, 40, "%d %b %y %H:%M %Z", tms);
      strcat(str, timestr);
    }
  else
    {
      strcat(str, "<h3>Lifetime</h3>Permanent\n");
    }
  
  

  strcat(str, "<h3>Media</h3><DL>");
  media=sess->media;
  while(media!=NULL)
    {
      char tmpstr[10];
      strcat(str, "<DT>");
      if (strcmp(media->media, "audio")==0)
	strcat(str, "<img src=/images/audio> ");
      else if (strcmp(media->media, "video")==0)
        strcat(str, "<img src=/images/video> ");
      else if (strcmp(media->media, "whiteboard")==0)
        strcat(str, "<img src=/images/whiteboard> ");
      else if (strcmp(media->media, "text")==0)
        strcat(str, "<img src=/images/text> ");
      strcat(str, media->media);
      strcat(str, "<DD> Address: ");
      htmlstrcat(str, inet_ntoa(media->addr));
      strcat(str, "<DD> Port: ");
      sprintf(tmpstr, "%u", media->port);
      strcat(str, tmpstr);
      strcat(str, "<DD> TTL: ");
      sprintf(tmpstr, "%u", media->ttl);
      strcat(str, tmpstr);
      strcat(str, "<DD> Protocol: ");
      htmlstrcat(str, media->proto);
      strcat(str, "<DD> Format: ");
      htmlstrcat(str, media->fmt);
      media=media->next;
    }
  strcat(str, "</DL>");
  return 0;
}


int html_print_session_name(char *str, hash_t aid)
{
  struct session *sess;
  struct media *media;
  sess=get_stored_session(aid, &entries);

  sprintf(str, "<h3><a href=/details/%u>", aid);
  htmlstrcat(str, sess->sname);
  strcat(str, "</a>");
  media=sess->media;
  while(media!=NULL)
    {
      if (strcmp(media->media, "audio")==0)
	strcat(str, "<img src=/images/audio> ");
      else if (strcmp(media->media, "video")==0)
        strcat(str, "<img src=/images/video> ");
      else if (strcmp(media->media, "whiteboard")==0)
        strcat(str, "<img src=/images/whiteboard> ");
      else if (strcmp(media->media, "text")==0)
        strcat(str, "<img src=/images/text> ");
      media=media->next;
    }
  strcat(str, "</h3>");
  return 0;
}


int html_print_session_summary(char *str, hash_t aid)
{
  struct session *sess;
  struct media *media;
  unsigned int origin;
  sess=get_stored_session(aid, &entries);
  origin=get_origin(aid, &entries);

  sprintf(str, "<h2><a href=/details/%u>", aid);
  htmlstrcat(str, sess->sname);
  strcat(str, "</a>");
  media=sess->media;
  while(media!=NULL)
    {
      if (strcmp(media->media, "audio")==0)
	strcat(str, "<img src=/images/audio> ");
      else if (strcmp(media->media, "video")==0)
        strcat(str, "<img src=/images/video> ");
      else if (strcmp(media->media, "whiteboard")==0)
        strcat(str, "<img src=/images/whiteboard> ");
      else if (strcmp(media->media, "text")==0)
        strcat(str, "<img src=/images/text> ");
      media=media->next;
    }
  strcat(str, "</h2>");
  htmlstrcat(str, sess->desc);
  switch(origin)
    {
    case UNKNOWN:
      strcat(str, "<br>UNKOWN origin");
      break;
    case LOCAL_WWW:
      strcat(str, "<br>Local WWW origin");
      break;
    case LOCAL_CLIENT:
      strcat(str, "<br>Local Sdr-client origin");
      break;
    case REMOTE:
      strcat(str, "<br>Remote origin");
      break;
    }
  return 0;
}

char htmlcharmap[256][9];
void setmap(char *s, int ix)
{
  strcpy(htmlcharmap[ix], s);
}
void inithtmlmap()
{
  int i;
  for(i=0;i<256;i++)
    {
      htmlcharmap[i][0]=(char)i;
      htmlcharmap[i][1]=(char)0;
    }
  setmap("&quot;", 34);
  setmap("&AElig;", 198);
  setmap("&Aacute;", 193);
  setmap("&Acirc;", 194);
  setmap("&Agrave;", 192);
  setmap("&Aring;", 197);
  setmap("&Atilde;", 195);
  setmap("&Auml;", 196);
  setmap("&Ccedil;", 199);
  setmap("&ETH;", 208);
  setmap("&Eacute;", 201);
  setmap("&Ecirc;", 202);
  setmap("&Egrave;", 200);
  setmap("&Euml;", 203);
  setmap("&Iacute;", 205);
  setmap("&Icirc;", 206);
  setmap("&Igrave;", 204);
  setmap("&Iuml;", 207);
  setmap("&Ntilde;", 209);
  setmap("&Oacute;", 211);
  setmap("&Ocirc;", 212);
  setmap("&Ograve;", 210);
  setmap("&Oslash;", 216);
  setmap("&Otilde;", 213);
  setmap("&Ouml;", 214);
  setmap("&THORN;", 222);
  setmap("&Uacute;", 218);
  setmap("&Ucirc;", 219);
  setmap("&Ugrave;", 217);
  setmap("&Uuml;", 220);
  setmap("&Yacute;", 221);
  setmap("&aacute;", 225);
  setmap("&acirc;", 226);
  setmap("&aelig;", 230);
  setmap("&agrave;", 224);
  setmap("&amp;", 38);
  setmap("&aring;", 229);
  setmap("&atilde;", 227);
  setmap("&auml;", 228);
  setmap("&ccedil;", 231);
  setmap("&eacute;", 233);
  setmap("&ecirc;", 234);
  setmap("&egrave;", 232);
  setmap("&eth;", 240);
  setmap("&euml;", 235);
  setmap("&gt;", 62);
  setmap("&iacute;", 237);
  setmap("&icirc;", 238);
  setmap("&igrave;", 236);
  setmap("&iuml;", 239);
  setmap("&lt;", 60);
#ifdef NOTDEF
  setmap("&nbsp;", 32);
#endif
  setmap("&ntilde;", 241);
  setmap("&oacute;", 243);
  setmap("&ocirc;", 244);
  setmap("&ograve;", 242);
  setmap("&oslash;", 248);
  setmap("&otilde;", 245);
  setmap("&ouml;", 246);
  setmap("&szlig;", 223);
  setmap("&thorn;", 254);
  setmap("&uacute;", 250);
  setmap("&ucirc;", 251);
  setmap("&ugrave;", 249);
  setmap("&uuml;", 252);
  setmap("&yacute;", 253);
  setmap("&yuml;", 255);
}


void htmlstrcat(char *str, char* new)
{
  char *is, *in;
  if ((new==NULL)||(str==NULL)) return;
  is=str+strlen(str);
  in=new;
  while(*in!='\0')
    {
      strcpy(is, htmlcharmap[(int)*in]);
      is+=strlen(htmlcharmap[(int)*in]);
      in++;
    }
}



