#include "sdr.h"
#include "sdr_server.h"
#include "store_entries.h"
#include "server_prototypes.h"
#include "server_scope.h"
#include "phone_xbm.h"
#include "mail_xbm.h"
#include "audio_xbm.h"
#include "video_xbm.h"
#include "whiteboard_xbm.h"
#include "text_xbm.h"
#define MAX_CONS 5
#define MAX_HTTP_REQUEST 2048
int httpsock;
int http_fd[MAX_CONS];

char *medianames[]={
"audio",
"video",
"whiteboard",
"text"};

extern unsigned int hostaddr;
extern struct entry_list entries;
struct entry_list new_sessions;
int latest_new_ad=0;

int next_free_fd();
int parse_http_request(int fd);
int parse_get_request(int fd, char *buf);
int parse_post_request(int fd, char *buf);
int parse_create_request(FILE *con, char *name, char *buf);
int send_http_header(FILE *con);
int send_home_page(FILE *con);
int send_session_listing(FILE *con, char *type);
int send_session_details(FILE *con, int aid);
int send_app_header(FILE *con);
int send_session_description(FILE *con, int aid);
int send_image_header(FILE *con);
int send_image(FILE *con, char *name);
int send_create_form(FILE * con, char *type);
int send_new_session_form(FILE *con);
char *geturlencline(char **buf);
void url_decode(char *s);
void setup_media(struct media* media, int mtype, char *line, struct scope *scope);
int httpport=7999;

int http_init()
{
  struct sockaddr_in addr;
  int i;
  char one=1;
  httpsock = socket(PF_INET, SOCK_STREAM, 6);
  setsockopt(httpsock, SOL_SOCKET, SO_REUSEADDR,
               (char *)&one, sizeof(one));
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons(httpport);
  while((  bind(httpsock, (struct sockaddr *)&addr, sizeof(addr) ))!=0)
    {
      perror("HTTP TCP bind failed");
      sleep(10);
    }
  if((listen(httpsock, MAX_CONS))!=0)
    {
      perror("Listen failed:");
      return -1;
    }
  for(i=0;i<MAX_CONS;i++)
    http_fd[i]=-1;
  return httpsock;
}

void http_closedown()
{
  close(httpsock);
}

int http_accept()
{
  static fd_set s_fdset;
  static struct timeval tv;
  int fdnum, addrlen;
  struct sockaddr_in addr;
  tv.tv_sec = 0;
  tv.tv_usec = 0;
  FD_ZERO(&s_fdset);
  FD_SET(httpsock, &s_fdset);
  select(httpsock+1, &s_fdset, 0, 0, &tv);
  if (FD_ISSET(httpsock, &s_fdset)!=0)
    {
      fprintf(stderr, "Connection arrived\n");
      addrlen=sizeof(addr);
      fdnum=next_free_fd();
      if((http_fd[fdnum] = accept(httpsock, (struct sockaddr *)&addr, &addrlen))<0)
	{
	  perror("Accept failed:");
	  return -1;
	}
    }
  return 0;
}

void http_fd_set(fd_set *fdset)
{
  int i;
  for(i=0;i<MAX_CONS;i++)
    {
      if(http_fd[i]>0)
	FD_SET(http_fd[i], fdset);
    }
}

void http_fd_isset(fd_set *fdset)
{
  int i;
  for(i=0;i<MAX_CONS;i++)
    {
      if(http_fd[i]>0)
	if (FD_ISSET(http_fd[i], fdset)!=0)
	  {
	    parse_http_request(http_fd[i]);
	    close(http_fd[i]);
	    http_fd[i]=-1;
	  }
    }
}

int next_free_fd()
{
  int i;
  for(i=0;i<MAX_CONS;i++)
    {
      if(http_fd[i]<0) return i;
    }
  perror("no free http fds left\n");
  abort();
}

int parse_http_request(int fd)
{
  char buf[MAX_HTTP_REQUEST];
  int length;
  length=read(fd, buf, MAX_HTTP_REQUEST);
  if(strncmp(buf, "GET ", 4)==0)
    {
      parse_get_request(fd, buf);
    }
  else if(strncmp(buf, "PUT ", 4)==0)
    {
      printf("Got a put request\n");
    }
  else if(strncmp(buf, "POST ", 5)==0)
    {
      printf("Got a post request\n->%s<-\n", buf);
      parse_post_request(fd, buf);
    }
  else if(strncmp(buf, "HEAD ", 5)==0)
    {
      printf("Got a head request\n");
    }
  else
    {
      fprintf(stderr, "Unknown http request\n");
    }
  return 0;
}

int parse_get_request(int fd, char *buf)
{
  FILE *con;
  char line [80], req[80], file[80], proto[80];
  con=fdopen(fd, "r+");
  printf("Got a get request\n");
  /*copy the line to protect the sscanf from overrunning its buffers*/
  strncpy(line, buf, 79);
  sscanf(line, "%s %s %s", req, file, proto);
  printf(">%s<>%s<>%s<\n", req, file, proto);
  if(strcmp(file, "/")==0)
    {
      if(strlen(proto)!=0)
	send_http_header(con);
      send_home_page(con);
    }
  else if(strncmp(file, "/listing/", 9)==0)
    {
      char *aidstr;
      if(strlen(proto)!=0)
	send_http_header(con);
      aidstr = file+9;
      send_session_listing(con, aidstr);
    }
  else if (strncmp(file, "/details/", 9)==0)
    {
      char *aidstr;
      int aid;
      if(strlen(proto)!=0)
	send_http_header(con);
      aidstr = file+9;
      aid=atoi(aidstr);
      send_session_details(con, aid);
    }
  else if (strncmp(file, "/launch/", 8)==0)
    {
      char *aidstr;
      int aid;
      send_app_header(con);
      aidstr = file+8;
      aid=atoi(aidstr);
      send_session_description(con, aid);
    }
  else if (strncmp(file, "/images/", 8)==0)
    {
      char *aidstr;
      aidstr = file+8;
      send_image_header(con);
      send_image(con, aidstr);
    }
  else if (strncmp(file, "/create/", 8)==0)
    {
      char *aidstr;
      aidstr = file+8;
      send_create_form(con, aidstr);
    }
  else
    {
      fprintf(con, "<h1>Sdr gateway error</h1>Sorry, didn't understand your request\n");
      fprintf(con, "<p>Try <a href=/>here</a> instead\n");
    }
  fclose(con);
  return 0;
}

int parse_post_request(int fd, char *buf)
{
  FILE *con;
  char line [80], req[80], file[80], proto[80];
  con=fdopen(fd, "r+");
  printf("Got a post request\n");
  /*copy the line to protect the sscanf from overrunning its buffers*/
  strncpy(line, buf, 79);
  sscanf(line, "%s %s %s", req, file, proto);
  printf(">%s<>%s<>%s<\n", req, file, proto);
  if(strncmp(file, "/create/", 8)==0)
    {
      parse_create_request(con, file+8, buf);
    }
  fclose(con);
  return 0;
}

int send_http_header(FILE *con)
{
  fprintf(con, "HTTP/1.0 200 OK\n");
  fprintf(con, "MIME-Version 1.0\n");
  fprintf(con, "Server: sdr_server/1.0\n");
  fprintf(con, "Content-Type: text/html\n\n");
  fprintf(con, "<title>Session Directory Gateway</title>\n");
  return 0;
}

int send_image_header(FILE *con)
{
  fprintf(con, "HTTP/1.0 200 OK\n");
  fprintf(con, "MIME-Version 1.0\n");
  fprintf(con, "Server: sdr_server/1.0\n");
  fprintf(con, "Content-Type: image/xbm\n\n");
  return 0;
}

int send_app_header(FILE *con)
{
  fprintf(con, "HTTP/1.0 200 OK\n");
  fprintf(con, "MIME-Version 1.0\n");
  fprintf(con, "Server: sdr_server/1.0\n");
  fprintf(con, "Content-Type: application/x-sdr\n\n");
  return 0;
}

int send_session_listing(FILE *con, char *type)
{
  hash_t *hashlist;
  int length, i;
  char str[1024];
  fprintf(con, "<h1>Sessions Currently Advertised</h1>\n");
  hashlist=create_hash_list(&length, &entries);
  if(strcmp(type, "normal")==0)
    {
      sort_hash_list(hashlist, length, ALPHABETIC);
      for(i=0;i<length;i++)
	{
	  html_print_session_summary(str, hashlist[i]);
	  fprintf(con, "%s", str);
	}
    }
  else if(strcmp(type, "short")==0)
    {
      sort_hash_list(hashlist, length, ALPHABETIC);
      for(i=0;i<length;i++)
	{
	  html_print_session_name(str, hashlist[i]);
	  fprintf(con, "%s", str);
	}
    }
  else if(strcmp(type, "bytime")==0)
    {
      sort_hash_list(hashlist, length, STARTTIME);
      for(i=0;i<length;i++)
	{
	  html_print_session_name(str, hashlist[i]);
	  fprintf(con, "%s", str);
	}
    }
  free_hash_list(hashlist);
  return 0;
}

int send_session_details(FILE *con, int aid)
{
  char str[4096];
  struct session *sess;
  fprintf(con, "<h1>Session Details</h1><hr>\n");
  sess=get_stored_session(aid, &entries);
  if(sess==NULL) return -1;
  html_print_session(sess, str);
  fprintf(con, "%s<hr>Your options are: <a href=/launch/%u>Join this session</a> or <a href=/>Return to session listing</a><hr>\n", str, sess->aid);
  return 0;
}
 
int send_session_description(FILE *con, int aid)
{
  char *str;
  int length;
  str=get_stored_entry(aid, &entries, &length);
  fwrite(str, length, 1, con);
  return 0;
}
 
int send_image(FILE *con, char *name)
{
  if(strcmp(name, "phone")==0)
    {
      fprintf(con, "%s\n", phone_xbm);
    }
  else if(strcmp(name, "email")==0)
    {
      fprintf(con, "%s\n", mail_xbm);
    }
  else if(strcmp(name, "audio")==0)
    {
      fprintf(con, "%s\n", audio_xbm);
    }
  else if(strcmp(name, "video")==0)
    {
      fprintf(con, "%s\n", video_xbm);
    }
  else if(strcmp(name, "whiteboard")==0)
    {
      fprintf(con, "%s\n", whiteboard_xbm);
    }
  else if(strcmp(name, "text")==0)
    {
      fprintf(con, "%s\n", text_xbm);
    }
  return 0;
}

int send_home_page(FILE *con)
{
  fprintf(con, "<H1>Welcome to the Session Directory Gateway</h1>
This is a gateway from the World Wide Web into the Mbone session directory.
<p>
Your options are:
<UL>
<LI>View a <a href=/listing/normal>summary of the sessions</a> currently advertised.
<LI>View a <a href=/listing/short> short listing of the sessions</a> currently advertised.
<LI>View a <a href=/listing/bytime> short listing of the sessions</a> currently advertised sorted by start time.
<LI>Create a <a href=/create/form>new session</a> and advertise it.
</UL>\n");
  return 0;
}

int send_create_form(FILE * con, char *type)
{
  if (strcmp(type, "form")==0)
    {
      send_http_header(con);
      send_new_session_form(con);
    }
  else if (strncmp(type, "advertise/", 10)==0)
    {
      unsigned int aid;
      aid=atoi(type+10);
      advertise_session(aid, &new_sessions);
    }
  else if (strncmp(type, "edit/", 5)==0)
    {
      unsigned int aid;
      aid=atoi(type+5);
#ifdef TBD
      edit_session(aid, &new_sessions);
#endif
    }
  return 0;
}
int send_new_session_form(FILE *con)
{
  char str[4096];
  create_form(str, 0);
  fprintf(con, str);
  return 0;
}

int parse_create_request(FILE *con, char *name, char *buf)
{
  char *line;
  char str[4096];
  char errorstr[4096];
  struct session *sess;
  struct timeval tv;
  int i;
  char namestr[256];
  struct scope *scope;

  printf("parse_creat_request %s\n", name);
  strcpy(errorstr, "");
  if (buf==NULL) return -1;
  while ((buf!=NULL)&&(buf[0]!='\n')&&(buf[0]!=(char)0x0d))
    {
      nextline(&buf);
    }
  nextline(&buf);
  line=buf;
  sess=malloc(sizeof(struct session));
  init_sess(sess);
  sess->username=storethisline("www-user");
  sess->orighost.s_addr=hostaddr;
#ifdef SYSV
  gettimeofday(&tv);
#else
  gettimeofday(&tv, NULL);
#endif
  sess->sessid=tv.tv_sec;
  sess->version=tv.tv_sec;

  /*Create a hash of originator data as advert ID*/
  sprintf(namestr, "%s%u%s", sess->username, sess->sessid, inet_ntoa(sess->orighost));
  sess->aid=0;
  for(i=0;i<=4;i++)
    namestr[strlen(namestr)+i]='\0';
  for(i=0;i<=strlen(namestr)/4;i++)
    {
      unsigned int tmp;
      bcopy(&namestr[i*4],(char *)&tmp,4);
      sess->aid=(sess->aid<<1|sess->aid>>31)^tmp;
    }

  while(line!=NULL)
    {
      char phone[80];
      char email[80];
      int sday, smon, syear, shour, smin, duration;
      char timestr[15];
      struct tm tms;
      struct media *lastmedia;
      line=geturlencline(&buf);
      if(line!=NULL)
	{
	  url_decode(line);
	  if(strncmp(line, "sname=", 6)==0)
	    {
	      if(strcmp(line+6, "")==0)
		strcat(errorstr, "No session name specified\n");	      
	      sess->sname=storethisline(line+6);
	      continue;
	    }
	  if(strncmp(line, "desc=", 5)==0)
	    {
	      if(strcmp(line+5, "")==0)
		strcat(errorstr, "No session description specified\n");
	      sess->desc=storethisline(line+5);
	      continue;
	    }
	  if(strncmp(line, "uri=", 4)==0)
	    {
	      if(strcmp(line+4, "")!=0)
		 sess->uri=storethisline(line+4);
	      continue;
	    }
	  if(strncmp(line, "name=", 5)==0)
	    {
	      if(strcmp(line+5, "")==0)
                strcat(errorstr, "No user name specified\n");
	      strncpy(phone, line+5, 80);
	      strncpy(email, line+5, 80);
	      continue;
	    }
	  if(strncmp(line, "phone=", 6)==0)
	    {
	      if(strcmp(line+6, "")==0)
                strcat(errorstr, "No phone number specified\n");
	      strcat(phone, " ");
	      strncat(phone, line+6, 80);
	      sess->phone=malloc(sizeof(struct lent));
	      sess->phone->data=storethisline(phone);
	      sess->phone->next=NULL;
	      sess->no_of_phones++;
	      continue;
	    }
	  if(strncmp(line, "email=", 6)==0)
	    {
	      if(strcmp(line+6, "")==0)
                strcat(errorstr, "No email address specified\n");
	      strcat(email, " <");
	      strncat(email, line+6, 80);
	      strcat(email, ">");
	      sess->email=malloc(sizeof(struct lent));
	      sess->email->data=storethisline(email);
	      sess->email->next=NULL;
	      sess->no_of_emails++;
	      continue;
	    }
	  if(strncmp(line, "scope=", 6)==0)
	    {
	      scope=lookup_scope(line+6);
	    }
	  if(strncmp(line, "rptmode=", 8)==0)
	    {
	    }
	  if(strncmp(line, "sday=", 5)==0)
	    {
	      sday=atoi(line+5);
	    }
	  if(strncmp(line, "smon=", 5)==0)
	    {
	      smon=atoi(line+5);
	    }
	  if(strncmp(line, "syear=", 6)==0)
	    {
	      syear=atoi(line+6);
	    }
	  if(strncmp(line, "stime=", 6)==0)
	    {
	      shour=atoi(line+6);
	      smin=atoi(line+9);
	    }
	  if(strncmp(line, "sdn=", 4)==0)
	    {
	      if(strncmp(line+4, "pm", 2)==0)
		shour+=12;
	      sprintf(timestr, "%d:%d:%d %d:%d", sday, smon, syear, shour, smin);
	      strptime(timestr, "%d:%m:%y %H:%M", &tms);
	      sess->starttime=mktime(&tms);
	    }
	  if(strncmp(line, "duration=", 9)==0)
	    {
	      duration=atoi(line+9);
	    }
	  if(strncmp(line, "dunits=", 7)==0)
	    {
	      if(strncmp("hours", line+7, 5)==0)
		duration*=3600;
	      else if (strncmp("days", line+7, 4)==0)
		duration*=86400;
	      sess->endtime=sess->starttime+duration;
	      sess->no_of_times++;
	      sess->time=malloc(sizeof(struct tent));
	      sess->time->starttime=sess->starttime;
	      sess->time->endtime=sess->endtime;
	    }
	  if(strncmp(line, "rptfor=", 7)==0)
	    {
	    }
	  if(strncmp(line, "rptunits=", 9)==0)
	    {
	    }
	  if(strncmp(line, "audio=", 6)==0)
	    {
	      if(strncmp(line, "audio=---", 9)!=0)
		{
		  struct media *tmp;
		  tmp=malloc(sizeof(struct media));
		  if(sess->media==NULL)
		    sess->media=tmp;
		  else
		    lastmedia->next=tmp;
		  lastmedia=tmp;
		  tmp->next=NULL;
		  sess->no_of_media++;
		  setup_media(tmp, AUDIO, line+6, scope);
		}
	    }
	  if(strncmp(line, "video=", 6)==0)
	    {
	      if(strncmp(line, "video=---", 9)!=0)
		{
		  struct media *tmp;
		  tmp=malloc(sizeof(struct media));
		  if(sess->media==NULL)
		      sess->media=tmp;
		  else
		    lastmedia->next=tmp;
		  lastmedia=tmp;
		  tmp->next=NULL;
		  sess->no_of_media++;
		  setup_media(tmp, VIDEO, line+6, scope);
		}
	    }
	  if(strncmp(line, "whiteboard=", 11)==0)
	    {
	      if(strncmp(line, "whiteboard=---", 14)!=0)
		{
		  struct media *tmp;
		  tmp=malloc(sizeof(struct media));
		  if(sess->media==NULL)
		      sess->media=tmp;
		  else
		    lastmedia->next=tmp;
		  lastmedia=tmp;
		  tmp->next=NULL;
		  sess->no_of_media++;
		  setup_media(tmp, WHITEBOARD, line+11, scope);
		}
	    }
	  if(strncmp(line, "text=", 5)==0)
	    {
	      if(strncmp(line, "text=---", 8)!=0)
		{
		  struct media *tmp;
		  tmp=malloc(sizeof(struct media));
		  if(sess->media==NULL)
		      sess->media=tmp;
		  else
		    lastmedia->next=tmp;
		  lastmedia=tmp;
		  tmp->next=NULL;
		  sess->no_of_media++;
		  setup_media(tmp, TEXT, line+5, scope);
		}
	    }
	}
    }
  send_http_header(con);
  if(strcmp(errorstr, "")!=0)
    {
      fprintf(con, "<h1>Error in Session Creation</h1>\n<PRE>%s</PRE>",
	      errorstr);
    }
  else
    {
      char *ad;
      ad=make_session_description(sess);
      store_entry(sess, ad, strlen(ad), &new_sessions, LOCAL_WWW, scope);
      fprintf(con, "<h1>Your new session will be:</h1>\n");
      html_print_session(sess, str);
      fprintf(con, str);
      fprintf(con, "<hr>Your Options are: \
<a href=/create/advertise/%u>Advertise this session</a>  \
<a href=/create/edit/%u>Edit this session</a>  \
<a href=/>Cancel</a><hr>\n", sess->aid, sess->aid);
    }
  return 0;
}


char *geturlencline(char **buf)
{
  char *tmp, *tmp2;
  tmp=strchr(*buf, '&');
  if(tmp==NULL) 
    tmp=strchr(*buf, '\r');
  if(tmp==NULL) 
    tmp=strchr(*buf, '\n');
  if(tmp==NULL)
    return NULL;
  *tmp='\0';
  tmp2=*buf;
  *buf=tmp+1;
  return tmp2;
}

void url_decode(char *buf)
{
  char *p;
  char h[3];

  if(buf==NULL) return;
  p=buf;
  h[2]='\0';
  while(*buf!='\0')
    {
      if(*buf=='+')
	*p=' ';
      else if(*buf=='%')
	{
	  h[0]=buf[1];
	  h[1]=buf[2];
	  *p=(char)strtol(h, NULL, 16);
	  buf+=2;
	}
      else
	*p=*buf;
      p++;buf++;
    }
  *p='\0';
}


void setup_media(struct media* media, int mtype, char *line, struct scope *scope)
{
  media->media=storethisline(medianames[mtype]);
  media->addr=generate_address(&(scope->baseaddr), scope->netmask);
  media->port=generate_port(media->media);
  media->ttl=scope->ttl;
  switch(mtype)
    {
    case AUDIO:
      {
	if(strncmp(line, "PCM", 3)==0)
	  media->fmt=storethisline("pcm");
	else if(strncmp(line, "DVI", 3)==0)
	  media->fmt=storethisline("dvi");
	else if(strncmp(line, "LPC", 3)==0)
	  media->fmt=storethisline("lpc");
	else if(strncmp(line, "GSM", 3)==0)
	  media->fmt=storethisline("gsm");
	media->proto=storethisline("rtp");
	break;
      }
    case VIDEO:
      {
	if(strncmp(line, "H261", 4)==0)
	  media->fmt=storethisline("h261");
	else if(strncmp(line, "nv", 3)==0)
	  media->fmt=storethisline("nv");
	media->proto=storethisline("rtp");
	break;
      }
    case WHITEBOARD:
      {
	if(strncmp(line, "wb", 2)==0)
	  media->fmt=storethisline("wb");
	else if(strncmp(line, "nt", 2)==0)
	  media->fmt=storethisline("nt");
	media->proto=storethisline("udp");
	break;
      }
    case TEXT:
      {
	if(strncmp(line, "mumble", 5)==0)
          media->fmt=storethisline("mumble");
	media->proto=storethisline("udp");
	break;
      }
    }
}
