#include "sdr.h"
#include "server_scope.h"

#define MAXLINELEN 200

void parse_scope_line(char *s);

int server_init(char *configfile)
{
  FILE *file;
  char ldata[MAXLINELEN];
  char *line;
  file=fopen(configfile, "r");
  if(file==NULL)
    {
      fprintf(stderr, "Couldn't open config file: %s\n", configfile);
      return -1;
    }
  while(!feof(file))
    {
      line=ldata;
      fgets(line, MAXLINELEN, file);
      while((line[0]==' ')||(line[0]=='\t'))
	line++;
      if(line[0]=='#')
	continue;
      if(strncmp(line, "scope", 5)==0)
	{
	  parse_scope_line(line+5);
	  continue;
	}
    }
  return 0;
}

void parse_scope_line(char *line)
{
  char base[20];
  char name[80];
  char *tmp;
  unsigned int port=9876;
  unsigned char ttl=1;
  unsigned int netmask=24;
  while(line[0]!='\0')
    {
      while((line[0]==' ')||(line[0]=='\t'))
	line++;
      if(strncmp(line, "base=", 5)==0)
	{
	  store_lump(base, line+5);
	}
      else if(strncmp(line, "ttl=", 4)==0)
	{
	  ttl=atoi(line+4);
	}
      else if(strncmp(line, "port=", 5)==0)
	{
	  port=atoi(line+5);
	}
      else if(strncmp(line, "netmask=", 8)==0)
	{
	  netmask=atoi(line+8);
	}
      else if(strncmp(line, "name=", 5)==0)
	{
	  store_quoted_lump(name, line+5);
	}
      else 
	{
	  fprintf(stderr, "Parse error while parsing scope definition: %s\n",
		  line);
	}
      tmp=strchr(line, ' ');
      if(tmp==NULL)
	tmp=strchr(line, '\t');
      if(tmp==NULL)
	line[0]='\0';
      else
	line=tmp;
    }
  add_admin_scope(name, ttl, port, base, netmask);
}

void store_lump(char *str, char *line)
{
  char *tmp;
  tmp=strchr(line, ' ');
  if(tmp==NULL)
    tmp=strchr(line, '\t');
  if(tmp==NULL)
    tmp=strchr(line, '\r');
  if(tmp==NULL)
    tmp=strchr(line, '\n');
  if(tmp==NULL)
    strcpy(str, line);
  else
    {
      strncpy(str, line, tmp-line);
      str[tmp-line]='\0';
    }
}

void store_quoted_lump(char *str, char *line)
{
  char *tmp=NULL;
  if(line[0]=='"') 
    {
      line++;
      tmp=strchr(line, '"');
    }
  if(tmp==NULL)
    tmp=strchr(line, ' ');
  if(tmp==NULL)
    tmp=strchr(line, '\t');
  if(tmp==NULL)
    tmp=strchr(line, '\r');
  if(tmp==NULL)
    tmp=strchr(line, '\n');
  if(tmp==NULL)
    strcpy(str, line);
  else
    {
      strncpy(str, line, tmp-line);
      str[tmp-line]='\0';
    }
}
