#include "sdr.h"
#include "sdr_server.h"
#include "store_entries.h"
#include "server_prototypes.h"
void free_lent_list(struct lent *l);
void free_tent_list(struct tent *l);
void free_media_list(struct media *l);
void free_string(char *c);

struct session *parse_session(char *ad)
{
  int ttl;
  char maddr[80];
  struct session *sess;
  static char namestr[MAXADSIZE];
  char creator[256], modtime[256], createtime[256], createaddr[256],
       in[256], ip[256];
  int i;


  sess=malloc(sizeof(struct session));
  init_sess(sess);
  if(ad==NULL) return NULL;
  if (strncmp(ad, "v=0", 3)!=0)
    {
      return NULL;
    }
  nextline(&ad);
  if(ad==NULL) return NULL;

  if(strncmp(ad, "o=", 2)!=0)
    {
      return NULL;
    }
  sscanf(ad+2, "%s %s %s %s %s %s", creator, createtime, modtime, in, ip,
           createaddr);
  if (check_net_type(in,ip)<0) return NULL;
 
  /*Create a hash of originator data as advert ID*/
  sprintf(namestr, "%s%s%s", creator, createtime, createaddr);
  sess->aid=0;
  for(i=0;i<=4;i++)
    namestr[strlen(namestr)+i]='\0';
  for(i=0;i<=strlen(namestr)/4;i++)
    {
      unsigned int tmp;
      bcopy(&namestr[i*4],(char *)&tmp,4);
      sess->aid=(sess->aid<<1|sess->aid>>31)^tmp;
    }

  nextline(&ad);
  if(ad==NULL) return NULL;

  if(strncmp(ad, "s=", 2)!=0)
    {
      return NULL;
    }
  ad+=2;
  sess->sname=storethisline(ad);
  nextline(&ad);
  if(ad==NULL) return NULL;

  if(strncmp(ad, "i=", 2)!=0)
    {
      return NULL;
    }
  ad+=2;
  sess->desc=storethisline(ad);
  nextline(&ad);
  if(ad==NULL) return NULL;
  if(strncmp(ad, "u=", 2)==0)
    {
      ad+=2;
      sess->uri=storethisline(ad);
      nextline(&ad);
    }

  while((strncmp(ad, "e=", 2)==0)||(strncmp(ad, "p=", 2)==0))
    {
      if(strncmp(ad, "e=", 2)==0)
	{
	  struct lent *tmp;
	  ad+=2;
	  sess->no_of_emails++;
	  tmp=sess->email;
	  sess->email=malloc(sizeof(struct lent));
	  sess->email->next=tmp;
	  sess->email->data=storethisline(ad);
	  nextline(&ad);
	}
      if(strncmp(ad, "p=", 2)==0)
	{
	  struct lent *tmp;
	  ad+=2;
	  sess->no_of_phones++;
	  tmp=sess->phone;
	  sess->phone=malloc(sizeof(struct lent));
	  sess->phone->next=tmp;
	  sess->phone->data=storethisline(ad);
	  nextline(&ad);
	}
    }
  if(strncmp(ad, "c=", 2)==0)
    {
      char net[10],proto[10], *ttlstr;
      ad+=2;
      sscanf(ad, "%s %s %s", net, proto, maddr);
      ttlstr=strchr(maddr, '/');
      *ttlstr='\0';
      ttlstr++;
      ttl=atoi(ttlstr);
      nextline(&ad);
    }
  while(strncmp(ad, "t=", 2)==0)
    {
      unsigned int t1, t2;
      struct tent *tmp;
      ad+=2;
      sscanf(ad, "%u %u", &t1, &t2);
      t1=ntptounix(t1);
      t2=ntptounix(t2);
      if((t1 < sess->starttime)||(sess->starttime==0))
	sess->starttime=t1;
      if(t2 > sess->endtime)
	sess->endtime=t2;
      tmp=sess->time;
      sess->time=malloc(sizeof(struct tent));
      sess->time->starttime=t1;
      sess->time->endtime=t2;
      sess->time->next=tmp;
      nextline(&ad);
      if(strncmp(ad, "z=", 2)==0)
	{
          nextline(&ad);
        }
    }
  
  

  while(strncmp(ad, "m=", 2)==0)
    {
      char media[80], fmt[80], proto[80], mmaddr[80];
      unsigned int port;
      struct media *tmp;
      int mttl=-1;
      ad+=2;
      sess->no_of_media++;
      tmp=sess->media;
      sess->media=malloc(sizeof(struct media));
      sess->no_of_attrs=0;
      sess->media->next=tmp;
      sscanf(ad, "%s %u %s %s", media, &port, proto, fmt);
      nextline(&ad);
      if(strncmp(ad, "c=", 2)==0)
	{
	  char net[10], prot[10], *ttlstr;
	  ad+=2;
	  sscanf(ad, "%s %s %s", net, prot, mmaddr);
	  ttlstr=strchr(mmaddr, '/');
	  *ttlstr='\0';
	  ttlstr++;
	  mttl=atoi(ttlstr);
	  nextline(&ad);
	}
      while(strncmp(ad, "a=", 2)==0)
	{
	  struct lent *tmp2;
	  ad+=2;
	  sess->media->no_of_attrs++;
	  tmp2=sess->media->attr;
	  sess->media->attr=malloc(sizeof(struct media));
	  sess->media->attr->next=tmp2;
	  sess->media->attr->data=storethisline(ad);
	  nextline(&ad);
	}
      if (mttl==-1)
	{
	  sess->media->addr.s_addr=inet_addr(maddr);
	  sess->media->ttl=ttl;
	}
      else
	{
	  sess->media->addr.s_addr=inet_addr(mmaddr);
	  sess->media->ttl=mttl;
	}
      sess->media->media=storethisline(media);
      sess->media->port=port;
      sess->media->proto=storethisline(proto);
      sess->media->fmt=storethisline(fmt);
    }
  return sess;
}

void nextline(char **ad)
{
  *ad=strchr(*ad, '\n');
  if(*ad==NULL) return;
  (*ad)++;
}

char *storethisline(char *ad)
{
  char *nl, *line;
  int len;

  nl=strchr(ad, '\n');
  if(nl!=NULL)
    len=(long)nl-(long)ad;
  else
    len=strlen(ad);
  line=malloc(len+1);
  strncpy(line, ad, len);
  line[len]='\0';
  return(line);
}

int extract_ttl(char *addrstr)
{
  char *ttlstr;
  ttlstr=strchr(addrstr, '/');
  if (ttlstr==NULL) return 0;
  *ttlstr='\0';
  return(atoi(ttlstr+1));
}

int check_net_type(char *in, char *ip)
{
  if (strncmp(in, "IN", 2)!=0)
    {
      fprintf(stderr, "sdr: expected network type IN, got %s\n",
	      in);
      return -1;
    }
  if (strncmp(ip, "IP4", 3)!=0)
    {
      fprintf(stderr, "sdr: expected address type IP4, got %s\n",
	      ip);
      return -1;
    }
  return 0;
}

void init_sess(struct session *sess)
{
  sess->sname=NULL;
  sess->desc=NULL;
  sess->username=NULL;
  sess->orighost.s_addr=0;
  sess->uri=NULL;
  sess->no_of_emails=0;
  sess->email=NULL;
  sess->no_of_phones=0;
  sess->phone=NULL;
  sess->no_of_times=0;
  sess->time=NULL;
  sess->no_of_attrs=0;
  sess->attr=NULL;
  sess->no_of_media=0;
  sess->media=NULL;
}

void free_session(struct session *sess)
{
  free_string(sess->username);
  free_string(sess->sname);
  free_string(sess->desc);
  free_string(sess->uri);
  free_lent_list(sess->email);
  free_lent_list(sess->phone);
  free_lent_list(sess->attr);
}

void free_lent_list(struct lent *l)
{
  struct lent *tmp1, *tmp2;
  tmp1=l;
  while(tmp1!=NULL)
    {
      tmp2=tmp1->next;
      free_string(tmp1->data);
      free(tmp1);
      tmp1=tmp2;
    }
}

void free_tent_list(struct tent *l)
{
  struct tent *tmp1, *tmp2;
  tmp1=l;
  while(tmp1!=NULL)
    {
      tmp2=tmp1->next;
      free(tmp1);
      tmp1=tmp2;
    }
}

void free_media_list(struct media *l)
{
  struct media *tmp1, *tmp2;
  tmp1=l;
  while(tmp1!=NULL)
    {
      tmp2=tmp1->next;
      free_string(tmp1->media);
      free_string(tmp1->proto);
      free_string(tmp1->fmt);
      free_lent_list(tmp1->attr);
      free(tmp1);
      tmp1=tmp2;
    }
}

void free_string(char *str)
{
  if(str!=NULL)
    free(str);
}


char *make_session_description(struct session *sess)
{
  char sdp[2048], tmp[2048], *rtn;
  struct media *media;
  struct lent *email;
  struct lent *phone;
  struct lent *attr;
  struct tent *time;
  int i, m;
  strcpy(sdp, "v=0\n");
  sprintf(tmp, "o=%s %u %u IN IP4 %s\n", sess->username,
	  sess->sessid, sess->version, inet_ntoa(sess->orighost));
  strcat(sdp, tmp);
  sprintf(tmp, "s=%s\n", sess->sname);
  strcat(sdp, tmp);
  sprintf(tmp, "i=%s\n", sess->desc);
  strcat(sdp, tmp);
  if(sess->uri!=NULL)
    {
      sprintf(tmp, "u=%s\n", sess->uri);
      strcat(sdp, tmp);
    }
  email=sess->email;
  for(i=0;i<sess->no_of_emails; i++)
    {
      sprintf(tmp, "e=%s\n", email->data);
      strcat(sdp, tmp);
      email=email->next;
    }
  phone=sess->phone;
  for(i=0;i<sess->no_of_phones; i++)
    {
      sprintf(tmp, "p=%s\n", phone->data);
      strcat(sdp, tmp);
      phone=phone->next;
    }
  time=sess->time;
  for(i=0;i<sess->no_of_times; i++)
    {
      sprintf(tmp, "t=%u %u\n", unixtontp(time->starttime), unixtontp(time->endtime));
      strcat(sdp, tmp);
      time=time->next;
    }
  attr=sess->attr;
  for(i=0;i<sess->no_of_attrs;i++)
    {
      sprintf(tmp, "a=%s\n", attr->data);
      strcat(sdp, tmp);
      attr=attr->next;
    }
  media=sess->media;
  for(m=0;m<sess->no_of_media;m++)
    {
      sprintf(tmp, "m=%s %u %s %s\n", media->media, media->port,
	      media->proto, media->fmt);
      strcat(sdp, tmp);

      sprintf(tmp, "c=IN IP4 %s/%u\n", inet_ntoa(media->addr), media->ttl);
      strcat(sdp, tmp);
      attr=media->attr;
      for(i=0;i<media->no_of_attrs;i++)
	{
	  sprintf(tmp, "a=%s\n", attr->data);
	  strcat(sdp, tmp);
	  attr=attr->next;
	}
      media=media->next;
    }
  rtn=malloc(strlen(sdp)+1);
  strcpy(rtn, sdp);
  return(rtn);
}
