#include "sdr.h"
#include "sdr_server.h"
#include "server_scope.h"
#include "store_entries.h"
#include "server_send.h"
#include "server_prototypes.h"

struct scope* first_scope=NULL;
struct scope* last_scope=NULL;
extern int rxsock[];
extern int no_of_rx_socks;
extern int txsock[];
extern int no_of_tx_socks;

unsigned long mask[]={
0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,
65535, 32767, 16383, 8191, 4095, 2047, 1023, 511,
255, 127, 63, 31, 15, 7, 3, 1, 0};

void add_admin_scope(char *name, unsigned char ttl, unsigned int port, 
		     char *baseaddr, unsigned int netmask)
{
  struct scope* tmp;
  struct in_addr addr;
  char *announce_addr;
  addr.s_addr=htonl(((unsigned long)(ntohl(inet_addr(baseaddr)))|mask[netmask])-1);
  announce_addr=inet_ntoa(addr);
  tmp=malloc(sizeof(struct scope));
  tmp->next=NULL;
  if(last_scope!=NULL)
    last_scope->next=tmp;
  last_scope=tmp;
  if(first_scope==NULL)
    first_scope=tmp;
  tmp->baseaddr.s_addr=inet_addr(baseaddr);
  if(netmask>32) 
    fprintf(stderr, "Illegal netmask in add_admin_scope\n");
  else
    tmp->netmask=netmask;
  if(port>65535) 
    fprintf(stderr, "Illegal port in add_admin_scope\n");
  else
    tmp->port=port;
  tmp->name=storethisline(name);
  tmp->ttl=ttl;
  printf("adding scope %s\n", tmp->name);
  sd_listen(announce_addr, port, rxsock, &no_of_rx_socks); 
  tmp->sock=sd_tx(announce_addr, port, ttl, txsock, &no_of_tx_socks); 
}

struct scope *lookup_scope(char *name)
{
  struct scope* tmp;
  tmp=first_scope;
  while(tmp!=NULL)
    {
      int len;
      /*note lookup does a first left-substring match so we can lookup
	stuff from the middle of a string we're parsing*/
      len=strlen(tmp->name);
      if(strncmp(tmp->name, name, len)==0)
	return tmp;
      tmp=tmp->next;
    }
  return NULL;
}

struct scope *find_scope(char *addr, unsigned char ttl)
{
  struct scope *tmp, *tmp2;
  unsigned int testaddr, tmpaddr;
  testaddr=inet_addr(addr);
  tmp=first_scope;
  /*pass 1*/
  while(tmp!=NULL)
    {
      if(((tmpaddr|mask[tmp->netmask])
	  ==(tmp->baseaddr.s_addr|mask[tmp->netmask]))
	 &&(ttl==tmp->ttl))
	return tmp;
      tmp=tmp->next;
    }
  /*pass 2 - no ttl matched - take whichever scope gives larger ttl*/
  printf("Warning: selecting a scope zone on approx match: %s/%d\n",
	 addr, ttl);
  tmp=first_scope;
  tmp2=NULL;
  while(tmp!=NULL)
    {
      if(((tmpaddr|mask[tmp->netmask])
	  ==(tmp->baseaddr.s_addr|mask[tmp->netmask]))
	 &&(ttl<=tmp->ttl))
      tmp2=tmp;
      tmp=tmp->next;
    }
  return tmp2;
}

char *get_scope_base_addr(struct scope* scope)
{

  return(inet_ntoa(scope->baseaddr));
}

struct scope *get_first_scope()
{
  return(first_scope);
}

struct scope *get_next_scope(struct scope *oldscope)
{
  return(oldscope->next);
}
