#include "sdr.h"
#include "sdr_server.h"
#include "store_entries.h"
#include "server_scope.h"
#include "server_send.h"
#include "server_prototypes.h"

struct session *first_session=NULL;
extern struct entry_list entries;
extern struct entry_list new_sessions;

struct send_queue rq;
#define TIMEOUT 20
/*MAXTIME just has to be longer than any other timer event*/
#define MAXTIME 0xffff

int advertise_session(hash_t aid)
{
  char *ad;
  struct session *sess;
  struct scope *scope;
  int adlen;
  u_long timeout;
  ad=get_stored_entry(aid, &new_sessions, &adlen);
  if(ad==NULL) return -1;
  sess=get_stored_session(aid, &new_sessions);
  scope=get_scope(aid, &entries);
  move_stored_entry(aid, &new_sessions, &entries);
  timeout=TIMEOUT+((drand48()/2.0)-0.25)*TIMEOUT;
  add_to_send_queue(aid, sess->endtime, timeout, scope, &rq);
  return 0;
}

void init_send_queue(struct send_queue *q)
{
  q->first_el=NULL;
  q->last_el=NULL;
}

void add_to_send_queue(hash_t aid, unsigned long endtime, 
		       unsigned long timeout, struct scope *scope,
		       struct send_queue *rq)
{
  struct send_el *sel;
  struct timeval tv;
#ifdef SYSV
  gettimeofday(&tv);
#else
  gettimeofday(&tv, NULL);
#endif
  sel=malloc(sizeof(struct send_el));
  sel->endtime=endtime;
  sel->aid=aid;
  sel->scope=scope;
  sel->nexttime=timeout+tv.tv_sec;
  if(rq->first_el==NULL)
    {
      sel->prev=NULL;
      sel->next=NULL;
      rq->first_el=sel;
      rq->last_el=sel;
      return;
    }
  rq->last_el->next=sel;
  sel->prev=rq->last_el;
  sel->next=NULL;
  rq->last_el=sel;
}

hash_t first_to_send(struct send_queue *rq)
{
  return rq->first_el->aid;
}

long time_to_send(struct send_queue *rq)
{
  struct timeval tv;
  if(rq->first_el==NULL)
    return MAXTIME;
#ifdef SYSV
  gettimeofday(&tv);
#else
  gettimeofday(&tv, NULL);
#endif
  return(rq->first_el->nexttime-tv.tv_sec);
}

void delete_el(struct send_el* sel, struct send_queue *rq)
{
  if(rq->first_el==NULL) return;
  if(rq->first_el==sel)
    {
      rq->first_el=sel->next;
      if(sel->next!=NULL)
	sel->next->prev=NULL;
    }
  if(rq->last_el==sel)
    {
      rq->last_el=sel->prev;
      if(sel->prev!=NULL)
	sel->prev->next=NULL;
    }
  if(sel->next!=NULL)
    if(sel->next->prev!=NULL)
      sel->next->prev=sel->prev;
  if(sel->prev!=NULL)
    if(sel->prev->next!=NULL)
      sel->prev->next=sel->next;
  free(sel);
  return;
}

void cycle_queue(struct send_queue *rq)
{
  struct send_el *sel;
  struct send_el *prev_el;
  struct timeval tv;
#ifdef SYSV
  gettimeofday(&tv);
#else
  gettimeofday(&tv, NULL);
#endif
  
  if(rq->first_el!=NULL)
    if((rq->first_el->endtime!=0)&&(rq->first_el->endtime <= tv.tv_sec))
      {
	delete_el(rq->first_el, rq);
	return;
      }
  if(rq->first_el==NULL) return;
  if(rq->first_el==rq->last_el) 
    {
      rq->first_el->nexttime=tv.tv_sec+TIMEOUT+((drand48()/2.0)-0.25)*TIMEOUT;
      return;
    }
  sel=rq->first_el;
  sel->nexttime=tv.tv_sec+TIMEOUT+((drand48()/2.0)-0.25)*TIMEOUT;
  if(rq->last_el->nexttime<sel->nexttime)
    {
      rq->first_el=sel->next;
      rq->first_el->prev=NULL;
      rq->last_el->next=sel;
      sel->prev=rq->last_el;
      rq->last_el=sel;
      sel->next=NULL;
    }
  else
    {
      prev_el=rq->last_el->prev;
      while(prev_el->nexttime>=sel->nexttime)
	{
	  prev_el=prev_el->prev;
	  if(prev_el==NULL) return;
	}
      rq->first_el=sel->next;
      rq->first_el->prev=NULL;
      sel->prev=prev_el;
      sel->next=prev_el->next;
      prev_el->next->prev=sel;
      prev_el->next=sel;
    }
}

void send_next_ad(struct send_queue *rq)
{
  char *ad;
  int len;
  struct scope *scope;
  if(rq->first_el==NULL) return;
  ad=get_stored_entry(rq->first_el->aid, &entries, &len);
  scope=get_scope(rq->first_el->aid, &entries);
  printf("**Sending ad: %s\n\n", ad);
  send_advert(ad, scope->sock, scope->ttl);
  cycle_queue(rq);
}
